#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Class about the diff.

:class:`DiffBase` defines the basic structure of a diff.

:class:`NotesDiff` defines the basic structure of a brief diff of notes.

:class:`CatalogDiff` defines the basic structure of a brief diff of catalog.

:class:`FileDiff` defines the basic structure of a brief diff of data file.

:class:`LabelDiff` defines the basic structure of a brief diff of data label.

:class:`SensorDiff` defines the basic structure of a brief diff of sensor.

:class:`DataDiff` defines the basic structure of a brief diff of data.

:class:`SegmentDiff` defines the basic structure of a brief diff of a segment.

:class:`DatasetDiff` defines the basic structure of a brief diff of a dataset.

"""

from typing import Any, Dict, Sequence, Tuple, Type, TypeVar, Union, overload

from tensorbay.client.lazy import PagingList
from tensorbay.utility import (
    AttrsMixin,
    NameMixin,
    ReprMixin,
    ReprType,
    SortedNameList,
    UserSequence,
    attr,
    camel,
    common_loads,
)


class DiffBase(AttrsMixin, ReprMixin):
    """This class defines the basic structure of a diff.

    Attributes:
        action: The concrete action.

    """

    _T = TypeVar("_T", bound="DiffBase")

    _repr_attrs: Tuple[str, ...] = ("action",)

    action: str = attr()

    def __init__(self, action: str) -> None:
        self.action = action

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`DiffBase` instance from the given contents.

        Arguments:
            contents: A dict containing all the information of the diff::

                    {
                        "action": <str>
                    }

        Returns:
            A :class:`DiffBase` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the information of the diff into a dict.

        Returns:
            A dict containing all the information of the diff::

                {
                    "action": <str>
                }

        """
        return self._dumps()


class NotesDiff(DiffBase):
    """This class defines the basic structure of a brief diff of notes."""


class CatalogDiff(DiffBase):
    """This class defines the basic structure of a brief diff of catalog."""


class FileDiff(DiffBase):
    """This class defines the basic structure of a brief diff of data file."""


class LabelDiff(DiffBase):
    """This class defines the basic structure of a brief diff of data label."""


class SensorDiff(DiffBase):
    """This class defines the basic structure of a brief diff of sensor."""


class DataDiff(DiffBase):
    """This class defines the basic structure of a diff statistic.

    Attributes:
        remote_path: The remote path.
        action: The action of data.
        file: The brief diff information of the file.
        label: The brief diff information of the labels.

    """

    _T = TypeVar("_T", bound="DataDiff")

    _repr_attrs = ("action", "file", "label")

    _repr_maxlevel = 2

    remote_path: str = attr(key=camel)
    file: FileDiff = attr()
    label: LabelDiff = attr()

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.remote_path}")'

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`DataDiff` instance from the given contents.

        Arguments:
            contents: A dict containing all the brief diff information of data::

                     {
                         "remotePath": <str>,
                         "action": <str>,
                         "file": {
                             "action": <str>
                         },
                         "label": {
                             "action": <str>
                         }
                     }

        Returns:
            A :class:`DataDiff` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the brief diff information of data into a dict.

        Returns:
            A dict containing all the brief diff information of data::

                 {
                     "remotePath": <str>,
                     "action": <str>,
                     "file": {
                         "action": <str>
                     },
                     "label": {
                         "action": <str>
                     }
                 }

        """
        return self._dumps()


class SegmentDiff(UserSequence[DataDiff], NameMixin):
    """This class defines the basic structure of a brief diff of a segment.

    Arguments:
        name: The segment name.
        action: The action of a segment.

    """

    _repr_attrs = ("action",)

    def __init__(self, name: str, action: str, data: PagingList[DataDiff]) -> None:
        super().__init__(name)

        self.action = action
        self._data = data


class DatasetDiff(Sequence[SegmentDiff], NameMixin):
    """This class defines the basic structure of a brief diff of a dataset.

    Arguments:
        name: The segment name.
        action: The action of a segment.

    """

    _repr_type = ReprType.SEQUENCE

    def __init__(self, name: str, segments: PagingList[SegmentDiff]) -> None:
        super().__init__(name)

        self._segments: SortedNameList[SegmentDiff] = SortedNameList()
        for segment in segments:
            self._segments.add(segment)

    def __len__(self) -> int:
        return self._segments.__len__()

    @overload
    def __getitem__(self, index: Union[int, str]) -> SegmentDiff:
        ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[SegmentDiff]:
        ...

    def __getitem__(
        self, index: Union[int, str, slice]
    ) -> Union[Sequence[SegmentDiff], SegmentDiff]:
        return self._segments.__getitem__(index)

    def __contains__(self, key: Any) -> bool:
        return self._segments.__contains__(key)



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the SCUT_FBP5500 dataset."""

from tensorbay.opendataset.SCUT_FBP5500.loader import SCUT_FBP5500

__all__ = ["SCUT_FBP5500"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of NeolixOD dataset."""

import os

from quaternion import from_rotation_vector

from tensorbay.dataset import Data, Dataset
from tensorbay.label import LabeledBox3D
from tensorbay.opendataset._utility import glob

DATASET_NAME = "NeolixOD"


def NeolixOD(path: str) -> Dataset:
    """`Neolix OD <https://gas.graviti.cn/dataset\
    /graviti-open-dataset/NeolixOD>`_ dataset.

    The file structure should be like::

        <path>
            bins/
                <id>.bin
            labels/
                <id>.txt
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    point_cloud_paths = glob(os.path.join(root_path, "bins", "*.bin"))

    for point_cloud_path in point_cloud_paths:
        data = Data(point_cloud_path)
        data.label.box3d = []

        point_cloud_id = os.path.basename(point_cloud_path)[:6]
        label_path = os.path.join(root_path, "labels", f"{point_cloud_id}.txt")

        with open(label_path, encoding="utf-8") as fp:
            for label_value_raw in fp:
                label_value = label_value_raw.rstrip().split()
                label = LabeledBox3D(
                    size=[float(label_value[10]), float(label_value[9]), float(label_value[8])],
                    translation=[
                        float(label_value[11]),
                        float(label_value[12]),
                        float(label_value[13]) + 0.5 * float(label_value[8]),
                    ],
                    rotation=from_rotation_vector((0, 0, float(label_value[14]))),
                    category=label_value[0],
                    attributes={
                        "Occlusion": int(label_value[1]),
                        "Truncation": bool(int(label_value[2])),
                        "Alpha": float(label_value[3]),
                    },
                )
                data.label.box3d.append(label)

        segment.append(data)
    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""ReprType and ReprMixin.

:class:`ReprType` is an enumeration type, which defines the repr strategy type and includes
'INSTANCE', 'SEQUENCE', 'MAPPING'.

:class:`ReprMixin` provides customized repr config and method.

"""

from abc import ABC
from enum import Enum, auto
from typing import Any, Callable, Dict, Iterable, Mapping, Sequence, Tuple, Type, TypeVar, Union


class ReprType(Enum):
    """ReprType is an enumeration type.

    It defines the repr strategy type and includes 'INSTANCE', 'SEQUENCE' and 'MAPPING'.

    """

    INSTANCE = auto()
    SEQUENCE = auto()
    MAPPING = auto()


class ReprMixin:
    """ReprMixin provides customized repr config and method."""

    _repr_type = ReprType.INSTANCE
    _repr_attrs: Iterable[str] = ()
    _repr_maxlevel = 1
    _repr_non_empty = False

    def __repr__(self) -> str:
        return _repr1(self, self._repr_maxlevel, self._repr_maxlevel, True)

    def __str__(self) -> str:
        return _repr1(self, self._repr_maxlevel, self._repr_maxlevel, False)

    def _repr_head(self) -> str:
        return self.__class__.__name__


class _ReprConfig:
    """Config for customized repr method.

    Arguments:
        indent: The indent width for different level repr string.
        maxlist: The max items displayed in the sequence repr string.
        maxdict: The max items displayed in the mapping repr string.

    """

    def __init__(self, indent: int = 2, maxlist: int = 16, maxdict: int = 16) -> None:
        self.indent = indent
        self.maxlist = maxlist
        self.maxdict = maxdict

    @property
    def indent(self) -> int:
        """The getter of repr indent width.

        Returns:
            The indent width of repr config.

        """
        return self._indent

    @indent.setter
    def indent(self, indent: int) -> None:
        """The setter of repr indent width.

        Arguments:
            indent: The width of the ident to be set.

        """
        self._indent = indent
        self._indent_str = " " * indent


repr_config = _ReprConfig()


# Explicitly inherit from 'abc.ABC' to tell pylint it is a abstract class
# Reference:
# - https://github.com/PyCQA/pylint/issues/3098
# - https://github.com/PyCQA/pylint/pull/3446
class _ReprSequence(ReprMixin, Sequence[Any], ABC):
    ...


class _ReprMapping(ReprMixin, Mapping[Any, Any], ABC):
    ...


_ReprPrinter = Callable[[Any, int, int, bool], str]

_PRINTERS: Dict[Union[Type[Any], ReprType], _ReprPrinter] = {}


class _PrinterRegister:
    """Decorator class to register repr printer functions to '_PRINTERS'.

    Arguments:
        key: The key of the '_PRINTERS' for repr dispatch.

    """

    _S = TypeVar("_S", bound=_ReprPrinter)

    def __init__(self, key: Union[Type[Any], ReprType]) -> None:
        self._key = key

    def __call__(self, printer: _S) -> _S:
        _PRINTERS[self._key] = printer
        return printer


def _repr1(obj: Any, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    printer_key = obj._repr_type if hasattr(obj, "_repr_type") else type(obj)
    printer = _PRINTERS.get(printer_key, None)
    return printer(obj, level, maxlevel, folding) if printer else repr(obj)


@_PrinterRegister(ReprType.INSTANCE)
def _repr_instance(obj: ReprMixin, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for `ReprType.INSTANCE`.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    return f"{obj._repr_head()}{_repr_attrs(obj, level, maxlevel, folding)}"


@_PrinterRegister(ReprType.SEQUENCE)
def _repr_sequence(obj: _ReprSequence, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for `ReprType.SEQUENCE`.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    return (
        f"{obj._repr_head()} {_repr_builtin_list(obj, level, maxlevel, folding)}"
        f"{_repr_attrs(obj, level, maxlevel, folding)}"
    )


@_PrinterRegister(ReprType.MAPPING)
def _repr_mapping(obj: _ReprMapping, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for `ReprType.MAPPING`.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    return (
        f"{obj._repr_head()} {_repr_builtin_dict(obj, level, maxlevel, folding)}"
        f"{_repr_attrs(obj, level, maxlevel, folding)}"
    )


@_PrinterRegister(list)
def _repr_builtin_list(obj: Sequence[Any], level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for buildin type list.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    return _repr_builtin_sequence(obj, level, maxlevel, folding, "[", "]")


@_PrinterRegister(tuple)
def _repr_builtin_tuple(obj: Sequence[Any], level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for buildin type tuple.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    return _repr_builtin_sequence(obj, level, maxlevel, folding, "(", ")")


@_PrinterRegister(dict)
def _repr_builtin_dict(obj: Mapping[Any, Any], level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for buildin type dict.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    if getattr(obj, "_repr_non_empty", False) and level <= 0:
        return "{...}"

    object_length = len(obj)
    if object_length == 0:
        return "{}"
    if level <= 0:
        return "{...}"
    newlevel = level - 1

    keys = tuple(obj)

    fold, unfold = _calculate_fold_number(object_length, repr_config.maxdict, folding)
    pieces = [
        f"{repr(key)}: {_repr1(obj[key], newlevel, maxlevel, folding)}" for key in keys[:unfold]
    ]
    if fold:
        key = keys[-1]
        pieces.append(f"... ({fold} items are folded)")
        pieces.append(f"{repr(key)}: {_repr1(obj[key], newlevel, maxlevel, folding)}")

    return _join_with_indent(pieces, level, maxlevel, "{", "}")


def _repr_attrs(obj: Any, level: int, maxlevel: int, folding: bool) -> str:
    """Customized repr method for object attributes.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.

    Returns:
        "repr" of the object.

    """
    # pylint: disable=protected-access
    if not obj._repr_attrs:
        return ""

    attributes = []
    for attribute in obj._repr_attrs:
        value = getattr(obj, attribute, ...)
        if value is not ...:
            attributes.append((attribute, value))

    if not attributes:
        return "()"

    if level <= 0:
        return "(...)"

    newlevel = level - 1
    pieces = (f"({key}): {_repr1(value, newlevel, maxlevel, folding)}" for key, value in attributes)

    return _join_with_indent(pieces, level, maxlevel, "(", ")")


def _repr_builtin_sequence(  # pylint: disable=too-many-arguments
    obj: Sequence[Any], level: int, maxlevel: int, folding: bool, left: str, right: str
) -> str:
    """Customized repr method for sequence.

    Arguments:
        obj: The object need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        folding: Whether fold the "repr" of the object.
        left: The left bracket symbol.
        right: The right bracket symbol.

    Returns:
        "repr" of the object.

    """
    if getattr(obj, "_repr_non_empty", False) and level <= 0:
        return f"{left}...{right}"

    object_length = len(obj)
    if object_length == 0:
        return f"{left}{right}"
    if level <= 0:
        return f"{left}...{right}"
    newlevel = level - 1

    fold, unfold = _calculate_fold_number(object_length, repr_config.maxlist, folding)
    pieces = [_repr1(item, newlevel, maxlevel, folding) for item in obj[:unfold]]
    if fold:
        pieces.append(f"... ({fold} items are folded)")
        pieces.append(_repr1(obj[-1], newlevel, maxlevel, folding))

    return _join_with_indent(pieces, level, maxlevel, left, right)


def _calculate_fold_number(length: int, max_length: int, folding: bool) -> Tuple[int, int]:
    """Calculate the number of the folded and unfolded items.

    Arguments:
        length: The object length.
        max_length: The configured max length.
        folding: Whether fold the "repr" of the object.

    Returns:
        The number of the folded items & the number of the unfolded items.

    """
    if folding and length > max_length:
        return length - max_length + 1, max_length - 2
    return 0, length


def _join_with_indent(
    pieces: Iterable[str], level: int, maxlevel: int, left: str, right: str
) -> str:
    """Handle the indent, newline and "," of the string list.

    Arguments:
        pieces: The string list need to be transferred to repr string.
        level: The current repr level.
        maxlevel: The max repr level.
        left: The left bracket symbol.
        right: The right bracket symbol.

    Returns:
        Repr string with indent, newline and ",".

    """
    # pylint: disable=protected-access
    inner_indent = repr_config._indent_str * (maxlevel - level + 1)
    outer_indent = repr_config._indent_str * (maxlevel - level)

    sep = f",\n{inner_indent}"
    return f"{left}\n{inner_indent}{sep.join(pieces)}\n{outer_indent}{right}"



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of CACD dataset."""

import os
from typing import Any, Dict, Tuple

from tensorbay.dataset import Data, Dataset
from tensorbay.exception import ModuleImportError
from tensorbay.label import CategoryInfo, Classification, ClassificationSubcatalog
from tensorbay.opendataset._utility import glob
from tensorbay.utility import NameList

DATASET_NAME = "CACD"
_ATTRIBUTES = {
    "attributes": [
        {
            "name": "name",
            "type": "string",
            "description": "celebrity name",
        },
        {
            "name": "age",
            "type": "number",
            "minimum": 14,
            "maximum": 62,
            "description": "Estimated age of the celebrity",
        },
        {
            "name": "birth",
            "type": "number",
            "minimum": 1951,
            "maximum": 1990,
            "description": "Celebrity birth year",
        },
        {
            "name": "range",
            "type": "number",
            "minimum": 1,
            "maximum": 50,
            "description": "Rank of the celebrity with same birth year in IMDB.com",
        },
        {
            "name": "year",
            "type": "number",
            "minimum": 2004,
            "maximum": 2013,
            "description": "Estimated year of which the photo was taken",
        },
        {
            "name": "lfw",
            "type": "boolean",
            "description": "Whether the celebrity is in LFW dataset.",
        },
    ]
}
_MAT_KEYS = ("name", "identity", "age", "birth", "lfw", "rank", "year")


def CACD(path: str) -> Dataset:
    """`Cross-Age Celebrity Dataset (CACD) <https://bcsiriuschen.github.io/CARC/>`_ dataset.

    The file structure should be like::

        <path>
            CACD2000/
                14_Aaron_Johnson_0001.jpg
                ...
            celebrity2000.mat

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(DATASET_NAME)
    dataset.catalog.classification = _get_subcatalog()
    segment = dataset.create_segment()
    image_files = glob(os.path.join(root_path, "CACD2000", "*.jpg"))
    labels_map = _get_labels_map(os.path.join(root_path, "celebrity2000.mat"))
    for image in image_files:
        category, attribute = labels_map[os.path.basename(image)]
        image_data = Data(image)
        image_data.label.classification = Classification(category, attribute)
        segment.append(image_data)
    return dataset


def _get_labels_map(path: str) -> Dict[str, Tuple[str, Dict[str, Any]]]:
    """Get celebrity_image_data from .mat file.

    Arguments:
        path: The root directory of the dataset.

    Raises:
        ModuleImportError: When the module "h5py" can not be found.

    Returns:
        A Dict of attributes.

    """
    try:
        from h5py import File  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name) from error

    mat_file = File(path, "r")
    celebrity_image_data = mat_file["celebrityImageData"]
    celebrity_data = mat_file["celebrityData"]

    # Name is a h5r object which can be searched in .mat file.
    id2name_map = {
        identity: _hdf5_to_str(mat_file[name])
        for identity, name in zip(celebrity_data["identity"][0], celebrity_data["name"][0])
    }
    labels_map = {}
    # The "name" is not the name of the celebrity but the name of the image file.
    for name, identity, *values in zip(*(celebrity_image_data[key][0] for key in _MAT_KEYS)):
        attribute = {"name": id2name_map[identity]}
        attribute.update(zip(_MAT_KEYS[2:], values))
        labels_map[_hdf5_to_str(mat_file[name])] = (str(int(identity)).zfill(4), attribute)
    return labels_map


def _get_subcatalog() -> ClassificationSubcatalog:
    categories: NameList[CategoryInfo] = NameList()
    for i in range(1, 2001):
        categories.append(CategoryInfo(str(i).zfill(4)))
    classification_subcatalog = ClassificationSubcatalog.loads(_ATTRIBUTES)
    classification_subcatalog.categories = categories
    return classification_subcatalog


def _hdf5_to_str(hdf5_string: Any) -> str:
    return "".join(chr(c[0]) for c in hdf5_string)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Vector, Vector2D, Vector3D.

:class:`Vector` is the base class of :class:`Vector2D` and :class:`Vector3D`. It contains the
coordinates of a 2D vector or a 3D vector.

:class:`Vector2D` contains the coordinates of a 2D vector, extending :class:`Vector`.

:class:`Vector3D` contains the coordinates of a 3D vector, extending :class:`Vector`.

"""

from itertools import zip_longest
from math import hypot, sqrt
from sys import version_info
from typing import Dict, Iterable, Mapping, Optional, Sequence, Tuple, Type, TypeVar, Union

from tensorbay.utility import ReprType, UserSequence

if version_info >= (3, 8):
    # math.hypot method supports n-dimensional points in Python3.8
    # only the two dimensional case was supported before Python3.8
    _hypot_for_n_dimension = hypot
else:

    def _hypot_for_n_dimension(*coordinates: float) -> float:  # type: ignore[misc]
        return sqrt(sum(x * x for x in coordinates))


_V = TypeVar("_V", bound="Vector")
_V2 = TypeVar("_V2", bound="Vector2D")
_V3 = TypeVar("_V3", bound="Vector3D")

_T = Union["Vector2D", "Vector3D"]


class Vector(UserSequence[float]):
    """This class defines the basic concept of Vector.

    :class:`Vector` contains the coordinates of a 2D vector or a 3D vector.

    Arguments:
        x: The x coordinate of the vector.
        y: The y coordinate of the vector.
        z: The z coordinate of the vector.

    Examples:
        >>> Vector(1, 2)
        Vector2D(1, 2)

        >>> Vector(1, 2, 3)
        Vector3D(1, 2, 3)

    """

    _data: Tuple[float, ...]

    _repr_type = ReprType.INSTANCE

    _DIMENSION: Optional[int] = None

    def __new__(  # type: ignore[misc]  # pylint: disable=unused-argument
        cls: Type[_V],
        x: float,
        y: float,
        z: Optional[float] = None,
    ) -> _T:
        """Create a new instance of :class:`Vector`.

        Arguments:
            x: The x coordinate of the vector.
            y: The y coordinate of the vector.
            z: The z coordinate of the vector.

        Returns:
            The created :class:`Vector2D` or :class:`Vector3D` object.

        """
        ReturnType = Vector2D if z is None else Vector3D

        obj: _T = object.__new__(ReturnType)
        return obj

    def __bool__(self) -> bool:
        return any(self._data)

    def __neg__(self: _V) -> _V:
        result: _V = object.__new__(self.__class__)
        result._data = tuple(-coordinate for coordinate in self._data)
        return result

    def __add__(self: _V, other: Iterable[float]) -> _V:
        """Calculate the sum of the vector and other vector.

        Arguments:
            other: The added vector.

        Returns:
            The sum vector of adding two vectors.

        """
        try:
            result: _V = object.__new__(self.__class__)
            result._data = tuple(i + j for i, j in zip_longest(self._data, other))
            return result
        except TypeError:
            return NotImplemented

    def __radd__(self: _V, other: Sequence[float]) -> _V:
        """Calculate the sum of the vector and the other vector.

        Arguments:
            other: The added vector.

        Returns:
            The sum vector of adding two vectors.

        """
        return self.__add__(other)

    def __sub__(self: _V, other: Iterable[float]) -> _V:
        try:
            result: _V = object.__new__(self.__class__)
            result._data = tuple(i - j for i, j in zip_longest(self._data, other))
            return result
        except TypeError:
            return NotImplemented

    def __rsub__(self: _V, other: Iterable[float]) -> _V:
        try:
            result: _V = object.__new__(self.__class__)
            result._data = tuple(i - j for i, j in zip_longest(other, self._data))
            return result
        except TypeError:
            return NotImplemented

    def __mul__(self: _V, other: float) -> _V:
        try:
            if isinstance(other, (int, float)):
                result: _V = object.__new__(self.__class__)
                result._data = tuple(i * other for i in self._data)
                return result
        except TypeError:
            pass

        return NotImplemented

    def __rmul__(self: _V, other: float) -> _V:
        return self.__mul__(other)

    def __truediv__(self: _V, other: float) -> _V:
        try:
            if isinstance(other, (int, float)):
                result: _V = object.__new__(self.__class__)
                result._data = tuple(i / other for i in self._data)
                return result
        except TypeError:
            pass

        return NotImplemented

    def __floordiv__(self: _V, other: float) -> _V:
        try:
            if isinstance(other, (int, float)):
                result: _V = object.__new__(self.__class__)
                result._data = tuple(i // other for i in self._data)
                return result
        except TypeError:
            pass

        return NotImplemented

    def __abs__(self) -> float:
        return _hypot_for_n_dimension(*self._data)

    def _repr_head(self) -> str:
        return f"{self.__class__.__name__}{self._data}"

    @staticmethod
    def loads(contents: Mapping[str, float]) -> _T:
        """Loads a :class:`Vector` from a dict containing coordinates of the vector.

        Arguments:
            contents: A dict containing coordinates of the vector.

        Returns:
            The loaded :class:`Vector2D` or :class:`Vector3D` object.

        Examples:
            >>> contents = {"x": 1.0, "y": 2.0}
            >>> Vector.loads(contents)
            Vector2D(1.0, 2.0)

            >>> contents = {"x": 1.0, "y": 2.0, "z": 3.0}
            >>> Vector.loads(contents)
            Vector3D(1.0, 2.0, 3.0)

        """
        if "z" in contents:
            return Vector3D.loads(contents)
        return Vector2D.loads(contents)


class Vector2D(Vector):
    """This class defines the concept of Vector2D.

    :class:`Vector2D` contains the coordinates of a 2D vector.

    Arguments:
        x: The x coordinate of the 2D vector.
        y: The y coordinate of the 2D vector.

    Examples:
        >>> Vector2D(1, 2)
        Vector2D(1, 2)

    """

    _DIMENSION = 2

    def __new__(  # pylint: disable=unused-argument
        cls: Type[_V2], *args: float, **kwargs: float
    ) -> _V2:
        """Create a :class:`Vector2D` instance.

        Arguments:
            args: The coordinates of the 2D vector.
            kwargs: The coordinates of the 2D vector.

        Returns:
            The created :class:`Vector2D` instance.

        """
        obj: _V2 = object.__new__(cls)
        return obj

    def __init__(self, x: float, y: float) -> None:
        self._data = (x, y)

    def __abs__(self) -> float:
        return hypot(*self._data)

    @classmethod
    def loads(cls: Type[_V2], contents: Mapping[str, float]) -> _V2:
        """Load a :class:`Vector2D` object from a dict containing coordinates of a 2D vector.

        Arguments:
            contents: A dict containing coordinates of a 2D vector.

        Returns:
            The loaded :class:`Vector2D` object.

        Examples:
            >>> contents = {"x": 1.0, "y": 2.0}
            >>> Vector2D.loads(contents)
            Vector2D(1.0, 2.0)

        """
        return cls(**contents)

    @property
    def x(self) -> float:
        """Return the x coordinate of the vector.

        Returns:
            X coordinate in float type.

        Examples:
            >>> vector_2d = Vector2D(1, 2)
            >>> vector_2d.x
            1

        """
        return self._data[0]

    @property
    def y(self) -> float:
        """Return the y coordinate of the vector.

        Returns:
            Y coordinate in float type.

        Examples:
            >>> vector_2d = Vector2D(1, 2)
            >>> vector_2d.y
            2

        """
        return self._data[1]

    def dumps(self) -> Dict[str, float]:
        """Dumps the vector into a dict.

        Returns:
            A dict containing the vector coordinate.

        Examples:
                >>> vector_2d = Vector2D(1, 2)
                >>> vector_2d.dumps()
                {'x': 1, 'y': 2}

        """
        return {"x": self._data[0], "y": self._data[1]}


class Vector3D(Vector):
    """This class defines the concept of Vector3D.

    :class:`Vector3D` contains the coordinates of a 3D Vector.

    Arguments:
        x: The x coordinate of the 3D vector.
        y: The y coordinate of the 3D vector.
        z: The z coordinate of the 3D vector.

    Examples:
        >>> Vector3D(1, 2, 3)
        Vector3D(1, 2, 3)

    """

    _DIMENSION = 3

    def __new__(  # pylint: disable=unused-argument
        cls: Type[_V3], *args: float, **kwargs: float
    ) -> _V3:
        """Create a :class:`Vector3D` instance.

        Arguments:
            args: The coordinates of the 3D vector.
            kwargs: The coordinates of the 3D vector.

        Returns:
            The created :class:`Vector3D` instance.

        """
        obj: _V3 = object.__new__(cls)
        return obj

    def __init__(self, x: float, y: float, z: float) -> None:
        self._data = (x, y, z)

    @classmethod
    def loads(cls: Type[_V3], contents: Mapping[str, float]) -> _V3:
        """Load a :class:`Vector3D` object from a dict containing coordinates of a 3D vector.

        Arguments:
            contents: A dict contains coordinates of a 3D vector.

        Returns:
            The loaded :class:`Vector3D` object.

        Examples:
            >>> contents = {"x": 1.0, "y": 2.0, "z": 3.0}
            >>> Vector3D.loads(contents)
            Vector3D(1.0, 2.0, 3.0)

        """
        return cls(**contents)

    @property
    def x(self) -> float:
        """Return the x coordinate of the vector.

        Returns:
             X coordinate in float type.

        Examples:
            >>> vector_3d = Vector3D(1, 2, 3)
            >>> vector_3d.x
            1

        """
        return self._data[0]

    @property
    def y(self) -> float:
        """Return the y coordinate of the vector.

        Returns:
            Y coordinate in float type.

        Examples:
            >>> vector_3d = Vector3D(1, 2, 3)
            >>> vector_3d.y
            2

        """
        return self._data[1]

    @property
    def z(self) -> float:
        """Return the z coordinate of the vector.

        Returns:
            Z coordinate in float type.

        Examples:
            >>> vector_3d = Vector3D(1, 2, 3)
            >>> vector_3d.z
            3

        """
        return self._data[2]

    def dumps(self) -> Dict[str, float]:
        """Dumps the vector into a dict.

        Returns:
            A dict containing the vector coordinates.

        Examples:
            >>> vector_3d = Vector3D(1, 2, 3)
            >>> vector_3d.dumps()
            {'x': 1, 'y': 2, 'z': 3}

        """
        return {"x": self._data[0], "y": self._data[1], "z": self._data[2]}



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""This file defines the glob method for open dataset."""

from glob import glob as buildin_glob
from typing import List

from tensorbay.exception import NoFileError


def glob(pathname: str, *, recursive: bool = False) -> List[str]:
    """Return a sorted list of paths matching a pathname pattern.

    The pattern may contain simple shell-style wildcards a la fnmatch.
    However, unlike fnmatch, filenames starting with a dot are special cases
    that are not matched by '*' and '?' patterns.

    Arguments:
        pathname: The pathname pattern.
        recursive: If recursive is true, the pattern '**' will match any files and
            zero or more directories and subdirectories.

    Returns:
        A sorted list of paths matching a pathname pattern.

    Raises:
        NoFileError: When there is no file matching the given pathname pattern.

    """
    paths = buildin_glob(pathname, recursive=recursive)
    if not paths:
        raise NoFileError(pattern=pathname)
    paths.sort()

    return paths



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Polygon.

:class:`Polygon` contains the coordinates of the vertexes of the polygon
and provides :meth:`Polygon.area` to calculate the area of the polygon.

"""

from typing import Dict, Iterable, List, Mapping, Optional, Sequence, Type, TypeVar

from tensorbay.geometry.point_list import MultiPointList2D, PointList2D
from tensorbay.geometry.vector import Vector2D
from tensorbay.utility import UserMutableSequence, common_loads


class Polygon(PointList2D[Vector2D]):
    """This class defines the concept of Polygon.

    :class:`Polygon` contains the coordinates of the vertexes of the polygon and provides
    :meth:`Polygon.area` to calculate the area of the polygon.

    Examples:
        >>> Polygon([[1, 2], [2, 3], [2, 2]])
        Polygon [
          Vector2D(1, 2),
          Vector2D(2, 3),
          Vector2D(2, 2)
        ]

    """

    _P = TypeVar("_P", bound="Polygon")

    _ElementType = Vector2D

    @classmethod
    def loads(cls: Type[_P], contents: Sequence[Mapping[str, float]]) -> _P:
        """Loads the information of :class:`Polygon`.

        Arguments:
            contents: A list of dictionary lists containing the coordinates
                of the vertexes of the polygon.

        Returns:
            The loaded :class:`Polygon` object.

        Examples:
            >>> contents = [{"x": 1.0, "y": 1.0}, {"x": 2.0, "y": 2.0}, {"x": 2.0, "y": 3.0}]
            >>> Polygon.loads(contents)
            Polygon [
              Vector2D(1.0, 1.0),
              Vector2D(2.0, 2.0),
              Vector2D(2.0, 3.0)
            ]

        """
        return common_loads(cls, contents)

    def area(self) -> float:
        """Return the area of the polygon.

        The area is positive if the rotating direction of the points is counterclockwise,
        and negative if clockwise.

        Returns:
            The area of the polygon.

        Examples:
            >>> polygon = Polygon([[1, 2], [2, 2], [2, 3]])
            >>> polygon.area()
            0.5

        """
        area = 0.0
        for i in range(len(self._data)):  # pylint: disable=consider-using-enumerate
            # pylint: disable=invalid-name
            x1, y1 = self._data[i - 1]
            x2, y2 = self._data[i]
            area += x1 * y2 - x2 * y1
        return area / 2


class MultiPolygon(MultiPointList2D[Polygon]):
    """This class defines the concept of MultiPolygon.

    :class:`MultiPolygon` contains a list of polygons.

    Arguments:
        polygons: A list of polygons.

    Examples:
        >>> MultiPolygon([[[1.0, 4.0], [2.0, 3.7], [7.0, 4.0]],
        ...               [[5.0, 7.0], [6.0, 7.0], [9.0, 8.0]]])
        MultiPolygon [
            Polygon [...]
            Polygon [...]
            ...
        ]

    """

    _P = TypeVar("_P", bound="MultiPolygon")
    _ElementType = Polygon

    def __init__(self, polygons: Optional[Iterable[Iterable[Iterable[float]]]]):
        super().__init__(polygons)

    @classmethod
    def loads(cls: Type[_P], contents: Sequence[Sequence[Mapping[str, float]]]) -> _P:
        """Loads a :class:`MultiPolygon` from the given contents.

        Arguments:
            contents: A list of dict lists containing
                the coordinates of the vertices of the polygon list.

        Returns:
            The loaded :class:`MultiPolyline2D` object.

        Examples:
            >>> contents = [[{'x': 1.0, 'y': 4.0}, {'x': 2.0, 'y': 3.7}, {'x': 7.0, 'y': 4.0}],
            ...             [{'x': 5.0, 'y': 7.0}, {'x': 6.0, 'y': 7.0}, {'x': 9.0, 'y': 8.0}]]
            >>> multipolygon = MultiPolygon.loads(contents)
            >>> multipolygon
            MultiPolygon [
                Polygon [...]
                Polygon [...]
                ...
            ]

        """
        return common_loads(cls, contents)

    def dumps(self) -> List[List[Dict[str, float]]]:
        """Dumps a :class:`MultiPolygon` into a polygon list.

        Returns:
            All the information of the :class:`MultiPolygon`.

        Examples:
            >>> multipolygon = MultiPolygon([[[1.0, 4.0], [2.0, 3.7], [7.0, 4.0]],
            ...                             [[5.0, 7.0], [6.0, 7.0], [9.0, 8.0]]])
            >>> multipolygon.dumps()
            [
                [{'x': 1.0, 'y': 4.0}, {'x': 2.0, 'y': 3.7}, {'x': 7.0, 'y': 4.0}],
                [{'x': 5,0, 'y': 7.0}, {'x': 6.0, 'y': 7.0}, {'x': 9.0, 'y': 8.0}]
            ]

        """
        return self._dumps()


class RLE(UserMutableSequence[int]):
    """This class defines the concept of RLE.

    :class:`RLE` contains an rle format mask.

    Arguments:
        rle: A rle format mask.

    Examples:
        >>> RLE([272, 2, 4, 4, 2, 9])
        RLE [
          272,
          2,
          ...
        ]

    """

    _data: List[int]

    def __init__(self, rle: Optional[Iterable[int]]):
        self._data = list(rle) if rle else []

    def _dumps(self) -> List[int]:
        return self._data

    def _loads(self, contents: List[int]) -> None:
        self._data = contents

    @classmethod
    def loads(cls: Type["RLE"], contents: List[int]) -> "RLE":
        """Loads a :class:RLE` from the given contents.

        Arguments:
            contents: One rle mask.

        Returns:
            The loaded :class:`RLE` object.

        Examples:
            >>> contents = [272, 2, 4, 4, 2, 9]
            >>> rle = RLE.loads(contents)
            >>> rle
            RLE [
              272,
              2,
              ...
            ]

        """
        return common_loads(cls, contents)

    def dumps(self) -> List[int]:
        """Dumps a :class:`RLE` into one rle mask.

        Returns:
            All the information of the :class:`RLE`.

        Examples:
            >>> rle = RLE([272, 2, 4, 4, 2, 9])
            >>> rle.dumps()
            [272, 2, 4, 4, 2, 9]

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of BioIDFace dataset."""

import os
from itertools import islice
from typing import List

from tensorbay.dataset import Data, Dataset
from tensorbay.geometry import Keypoint2D
from tensorbay.label import LabeledKeypoints2D
from tensorbay.opendataset._utility import glob

DATASET_NAME = "BioIDFace"


def BioIDFace(path: str) -> Dataset:
    """`BioID Face <https://www.bioid.com/facedb/>`_ dataset.

    The folder structure should be like::

                <path>
                    BioID-FaceDatabase-V1.2/
                        BioID_0000.eye
                        BioID_0000.pgm
                        ...
                    points_20/
                        bioid_0000.pts

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))
    segment = dataset.create_segment()

    image_paths = glob(os.path.join(root_path, "BioID-FaceDatabase-V1.2", "*.pgm"))
    face_keypoints_paths = glob(os.path.join(root_path, "points_20", "*.pts"))

    for image_path, face_keypoints_path in zip(image_paths, face_keypoints_paths):
        data = Data(image_path)
        data.label.keypoints2d = _get_label(
            f"{os.path.splitext(image_path)[0]}.eye", face_keypoints_path
        )

        segment.append(data)

    return dataset


def _get_label(eye_keypoints_path: str, face_keypoints_path: str) -> List[LabeledKeypoints2D]:
    eye_keypoints = LabeledKeypoints2D(category="EyePosition")
    with open(eye_keypoints_path, "r", encoding="utf-8") as fp:
        fp.readline()  # The first line is like: #LX     LY      RX      RY
        lx, ly, rx, ry = map(int, fp.readline().split())
        eye_keypoints.append(Keypoint2D(lx, ly))
        eye_keypoints.append(Keypoint2D(rx, ry))

    face_keypoints = LabeledKeypoints2D(category="Face")
    with open(face_keypoints_path, "r", encoding="utf-8") as fp:
        # The annotation file is like:
        #  1 version: 1
        #  2 n_points: 20
        #  3 {
        #  4 159.128 108.541
        #    ...
        # 24 }
        for line in islice(fp, 3, 23):
            x, y = map(float, line.split())
            face_keypoints.append(Keypoint2D(x, y))

    return [eye_keypoints, face_keypoints]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloaders of DAVIS2017SemiSupervised dataset and DAVIS2017Unsupervised dataset."""

import os
from itertools import product
from typing import Callable, Dict, Iterator

from tensorbay.dataset import Data, Dataset
from tensorbay.label import InstanceMask, SemanticMask
from tensorbay.opendataset._utility import glob

_RESOLUTIONS = ("480p", "Full-Resolution")
_SEMI_SUPERVISED_DATASET_NAME = "DAVIS2017SemiSupervised"
_UNSUPERVISED_DATASET_NAME = "DAVIS2017Unsupervised"


def DAVIS2017SemiSupervised(path: str) -> Dataset:
    """`DAVIS2017SemiSupervised <https://davischallenge.org/davis2017/code.html>`_ dataset.

    The file structure should be like::

        <path>
            Annotations/
                480p/
                    aerobatics/
                        00000.png
                    bear/
                    ...
                Full-Resolution/
                    aerobatics/
                        00000.png
                    bear/
                    ...
            Annotations_semantics/
                480p/
                    aerobatics/
                        00000.png
                    bear/
                    ...
                Full-Resolution/
                    aerobatics/
                        00000.png
                    bear/
                    ...
            JPEGImages/
                480p/
                    aerobatics/
                        00000.jpg
                        00001.jpg
                        ...
                    bear/
                    ...
                Full-Resolution/
                    aerobatics/
                        00000.jpg
                        00001.jpg
                        ...
                    bear/
                    ...
            ImageSets/
                2017/
                    test-challenge.txt
                    test-dev.txt
                    train.txt
                    val.txt

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(_SEMI_SUPERVISED_DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_semi_supervised.json"))

    data_getters: Dict[str, Callable[..., Data]] = {
        "train.txt": _get_semi_supervised_labeled_data,
        "val.txt": _get_semi_supervised_labeled_data,
        "test-dev.txt": lambda image_path, *args: Data(image_path),
        "test-challenge.txt": lambda image_path, *args: Data(image_path),
    }

    for file_path, data_getter in data_getters.items():
        for resolution, segment_name in product(
            _RESOLUTIONS,
            _generate_segment_name(os.path.join(root_path, "ImageSets", "2017", file_path)),
        ):
            stem = os.path.splitext(file_path)[0]
            segment = dataset.create_segment(f"{stem}_{segment_name}_{resolution}")
            image_paths = glob(
                os.path.join(root_path, "JPEGImages", resolution, segment_name, "*.jpg")
            )
            first_image_path = image_paths.pop(0)
            #  In the test segment of the semi_supervised task, only the mask of the first image
            #  will be given.
            segment.append(
                _get_semi_supervised_labeled_data(
                    first_image_path, root_path, resolution, segment_name
                )
            )
            for image_path in image_paths:
                segment.append(data_getter(image_path, root_path, resolution, segment_name))
    return dataset


def DAVIS2017Unsupervised(path: str) -> Dataset:
    """`DAVIS2017Unsupervised <https://davischallenge.org/davis2017/code.html>`_ dataset.

    The file structure should be like::

        <path>
            Annotations_unsuperviseds/
                480p/
                    bear/
                        00000.png
                        00001.png
                        ...
                    ...
                Full-Resolution/
                    bear/
                        00000.png
                        00001.png
                        ...
                    ...
            JPEGImages/
                480p/
                    aerobatics/
                        00000.jpg
                        00001.jpg
                        ...
                    bear/
                        00000.jpg
                        00001.jpg
                        ...
                    ...
                Full-Resolution/
                    aerobatics/
                        00000.jpg
                        00001.jpg
                        ...
                    bear/
                        00000.jpg
                        00001.jpg
                        ...
                    ...
            ImageSets/
                2017/
                    train.txt
                    val.txt
                2019/
                    test-challenge.txt
                    test-dev.txt

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class: `~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))
    dataset = Dataset(_UNSUPERVISED_DATASET_NAME)
    dataset.notes.is_continuous = True
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog_unsupervised.json"))

    data_getters: Dict[str, Callable[..., Data]] = {
        os.path.join("2017", "train.txt"): _get_unsupervised_labeled_data,
        os.path.join("2017", "val.txt"): _get_unsupervised_labeled_data,
        os.path.join("2019", "test-dev.txt"): lambda image_path, *args: Data(image_path),
        os.path.join("2019", "test-challenge.txt"): lambda image_path, *args: Data(image_path),
    }

    for file_path, data_getter in data_getters.items():
        for resolution, segment_name in product(
            _RESOLUTIONS,
            _generate_segment_name(os.path.join(root_path, "ImageSets", file_path)),
        ):
            stem = os.path.splitext(os.path.basename(file_path))[0]
            segment = dataset.create_segment(f"{stem}_{segment_name}_{resolution}")
            for image_path in glob(
                os.path.join(root_path, "JPEGImages", resolution, segment_name, "*.jpg")
            ):
                segment.append(data_getter(image_path, root_path, resolution, segment_name))
    return dataset


def _generate_segment_name(file_path: str) -> Iterator[str]:
    with open(file_path, "r", encoding="utf-8") as fp:
        for segment_name in fp:
            yield segment_name.strip()


def _get_semi_supervised_labeled_data(
    image_path: str, root_path: str, resolution: str, segment_name: str
) -> Data:
    data = Data(image_path)
    label = data.label
    mask_stem = os.path.splitext(os.path.basename(data.path))[0]
    mask_path = os.path.join(resolution, segment_name, f"{mask_stem}.png")

    label.instance_mask = InstanceMask(os.path.join(root_path, "Annotations", mask_path))
    label.semantic_mask = SemanticMask(os.path.join(root_path, "Annotations_semantics", mask_path))
    return data


def _get_unsupervised_labeled_data(
    image_path: str, root_path: str, resolution: str, segment_name: str
) -> Data:
    data = Data(image_path)
    label = data.label
    mask_stem = os.path.splitext(os.path.basename(data.path))[0]
    mask_path = os.path.join(resolution, segment_name, f"{mask_stem}.png")

    label.instance_mask = InstanceMask(
        os.path.join(root_path, "Annotations_unsupervised", mask_path)
    )
    return data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of CIHP dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.label import InstanceMask, SemanticMask

DATASET_NAME = "CIHP"
_SEGMENTS_INFO = {"train": "Training", "val": "Validation", "test": "Testing"}


def CIHP(path: str) -> Dataset:
    """`CIHP <https://github.com/Engineering-Course/CIHP_PGN>`_ dataset.

    The file structure should be like::

        <path>
            Testing/
                Images/
                    0000002.jpg
                    ...
                test_id.txt
            Training/
                Images/
                    0000006.jpg
                    ...
                Category_ids/
                    0000006.png
                    ...
                Instance_ids/
                    0000006.png
                    ...
                train_id.txt
            Validation/
                Images/
                    0000001.jpg
                    ...
                Category_ids/
                    0000001.png
                    ...
                Instance_ids/
                    0000001.png
                    ...
                val_id.txt

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.join(
        os.path.abspath(os.path.expanduser(path)), "instance-level_human_parsing"
    )

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for segment_name, segment_path in _SEGMENTS_INFO.items():
        segment = dataset.create_segment(segment_name)
        segment_abspath = os.path.join(root_path, segment_path)
        image_path = os.path.join(segment_abspath, "Images")
        with open(
            os.path.join(segment_abspath, f"{segment_name}_id.txt"), "r", encoding="utf-8"
        ) as fp:
            if segment_name == "test":
                for stem in fp:
                    segment.append(Data(os.path.join(image_path, f"{stem.rstrip()}.jpg")))
            else:
                category_path = os.path.join(segment_abspath, "Category_ids")
                instance_path = os.path.join(segment_abspath, "Instance_ids")
                for stem in fp:
                    stem = stem.rstrip()
                    data = Data(os.path.join(image_path, f"{stem}.jpg"))
                    label = data.label
                    png_filename = f"{stem}.png"
                    label.semantic_mask = SemanticMask(os.path.join(category_path, png_filename))
                    label.instance_mask = InstanceMask(os.path.join(instance_path, png_filename))
                    segment.append(data)
    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Healthcheck related methods.

:meth:`healthcheck` finds all errors in the :class:`~tensorbay.dataset.dataset.Dataset`
or :class:`~tensorbay.dataset.dataset.FusionDataset`,
including basic errors and catalog errors.

"""

from typing import Union

from tensorbay.dataset import Dataset, FusionDataset
from tensorbay.healthcheck.basic_check import check_basic
from tensorbay.healthcheck.catalog_check import check_catalog
from tensorbay.healthcheck.report import HealthReport


def healthcheck(dataset: Union[Dataset, FusionDataset]) -> HealthReport:
    """Healthcheck for Dataset or FusionDataset.

    Arguments:
        dataset: The :class:`~tensorbay.dataset.dataset.Dataset`
            or :class:`~tensorbay.dataset.dataset.FusionDataset` for healthchecking.

    Returns:
        The full result of the healthcheck which contains all errors found.

    """
    report = HealthReport()

    with report.basic_reports as basic_reports:
        for basic_error in check_basic(dataset):
            basic_reports.append(basic_error)

    with report.subcatalog_reports as subcatalog_reports:
        for label_type, attribute_info_error in check_catalog(dataset.catalog):
            subcatalog_reports[label_type].append(attribute_info_error)

    return report



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# type: ignore
# pylint: disable=redefined-outer-name

"""Pytest fixture config."""

import sys

import pytest

from tensorbay.client import gas, segment, version
from tensorbay.client.gas import DEFAULT_BRANCH
from tensorbay.client.struct import Branch, Draft
from tensorbay.client.tests.utility import mock_response
from tensorbay.dataset import RemoteData
from tensorbay.exception import UnauthorizedError


def function_fixture(fixture):
    """A decorator that makes fixture only run when it is called.

    Arguments:
        fixture: The fixture.

    Returns:
        The wrapped fixture.

    """

    @pytest.fixture
    def wrapper():
        return fixture

    return wrapper


@function_fixture
def mock_list_datasets(mocker):
    """Mock the listDatasets OpenAPI.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = {
        "datasets": [
            {
                "id": "123456",
                "name": "test",
                "type": 0,
                "defaultBranch": DEFAULT_BRANCH,
                "updateTime": 1622530298,
                "owner": "",
            }
        ],
        "offset": 0,
        "recordSize": 1,
        "totalCount": 1,
    }
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do",
            return_value=mock_response(data=response_data),
        ),
        response_data,
    )


@function_fixture
def mock_get_dataset(mocker, is_fusion, is_public):
    """Mock the _get_dataset funcion of GAS class.

    Arguments:
        mocker: The mocker fixture.
        is_fusion: Whether the dataset is a fusion dataset.
        is_public: Whether the dataset is a public dataset.

    Returns:
        The patched mocker and response data.

    """
    response_data = {
        "id": "123456",
        "type": int(is_fusion),
        "commitId": "4",
        "defaultBranch": DEFAULT_BRANCH,
        "alias": "alias",
        "isPublic": is_public,
    }
    return (
        mocker.patch(
            f"{gas.__name__}.GAS._get_dataset",
            return_value=response_data,
        ),
        response_data,
    )


@function_fixture
def mock_list_segments(mocker):
    """Mock the listSegments OpenAPI.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = {
        "offset": 0,
        "recordSize": 2,
        "totalCount": 2,
        "segments": [
            {"name": "segment0", "description": ""},
            {"name": "segment1", "description": ""},
        ],
    }
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do",
            return_value=mock_response(data=response_data),
        ),
        response_data,
    )


@function_fixture
def mock_paths(mocker):
    """Mock the list_data_paths funcion of SegmentClient class.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = [f"data{i}.png" for i in range(2)]
    return (
        mocker.patch(
            f"{segment.__name__}.SegmentClient.list_data_paths",
            return_value=response_data,
        ),
        response_data,
    )


@function_fixture
def mock_get_data(mocker, remote_path):
    """Mock the get_data funcion of SegmentClient class.

    Arguments:
        mocker: The mocker fixture.
        remote_path: The remote path of data.

    Returns:
        The patched mocker and response data.

    """
    response_data = [RemoteData(remote_path=remote_path)]
    return (
        mocker.patch(
            f"{segment.__name__}.SegmentClient.get_data",
            return_value=response_data,
        ),
        response_data,
    )


@function_fixture
def mock_create_dataset(mocker):
    """Mock the createDataset OpenAPI.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = {"id": "123456"}
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do", return_value=mock_response(data=response_data)
        ),
        response_data,
    )


@function_fixture
def mock_delete_dataset(mocker, is_fusion, is_public, mock_get_dataset):
    """Mock the deleteDataset OpenAPI.

    Arguments:
        mocker: The mocker fixture.
        is_fusion: Whether the dataset is a fusion dataset.
        is_public: Whether the dataset is a public dataset.
        mock_get_dataset: Mock the _get_dataset method of GAS class.

    Returns:
        The patched mocker and response data.

    """
    _, response_data = mock_get_dataset(mocker, is_fusion, is_public)
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do",
            return_value=response_data,
        ),
        response_data,
    )


@function_fixture
def mock_list_drafts(mocker, branch_name):
    """Mock the list_drafts method of VersionControlMixin class.

    Arguments:
        mocker: The mocker fixture.
        branch_name: The branch name of draft.

    Returns:
        The patched mocker and response data.

    """
    drafts = [
        {
            "number": 1,
            "title": "draft1",
            "branchName": branch_name,
            "status": "OPEN",
            "parentCommitId": "4c564ea07f4e47679ec8c63d238bb3a1",
            "author": {"name": "draft author", "date": 1636967807},
            "updatedAt": 1636967807,
            "description": "first draft of test_draft",
        },
        {
            "number": 2,
            "title": "",
            "branchName": branch_name,
            "status": "CLOSED",
            "parentCommitId": "4c564ea07f4e47679ec8c63d238bb3a1",
            "author": {"name": "draft author", "date": 1636967807},
            "updatedAt": 1636967807,
            "description": "",
        },
    ]
    response = [Draft.loads(draft_response) for draft_response in drafts]
    return (
        mocker.patch(f"{version.__name__}.VersionControlMixin.list_drafts", return_value=response),
        response,
    )


@function_fixture
def mock_get_branch(mocker, commit_id):
    """Mock the get_branch method of VersionControlMixin class.

    Arguments:
        mocker: The mocker fixture.
        commit_id: The commit id of the branch.

    Returns:
        The patched mocker and response data.

    """
    response_data = Branch.loads(
        {
            "name": DEFAULT_BRANCH,
            "commitId": commit_id,
            "parentCommitId": "",
            "title": "test_draft",
            "description": "",
            "committer": {"name": "", "date": 1632454975},
        }
    )
    return (
        mocker.patch(
            f"{version.__name__}.VersionControlMixin.get_branch", return_value=response_data
        ),
        response_data,
    )


@function_fixture
def mock_update_draft(mocker):
    """Mock the updateDraft OpenAPI.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = {"draftNumber": 1}
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do", return_value=mock_response(data=response_data)
        ),
        response_data,
    )


@function_fixture
def mock_create_draft(mocker):
    """Mock the createDraft OpenAPI.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = {"draftNumber": 1}
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do", return_value=mock_response(data=response_data)
        ),
        response_data,
    )


@function_fixture
def mock_close_draft(mocker):
    """Mock the closeDraft OpenAPI.

    Arguments:
        mocker: The mocker fixture.

    Returns:
        The patched mocker and response data.

    """
    response_data = {"status": "CLOSED"}
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do", return_value=mock_response(data=response_data)
        ),
        response_data,
    )


@function_fixture
def mock_list_data_details(mocker, num: int = 1):
    """Mock the getDataDetails openAPI.

    Arguments:
        mocker: The mocker fixture.
        num: The number of data.

    Returns:
        The patched mocker and response data.

    """
    response_data = {
        "dataDetails": [
            {
                "remotePath": f"data{i}.png",
                "timestamp": 1614667532,
                "label": {},
                "url": "url",
            }
            for i in range(num)
        ],
        "offset": 0,
        "recordSize": num,
        "totalCount": num,
    }
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do",
            return_value=mock_response(data=response_data),
        ),
        response_data,
    )


@function_fixture
def mock_get_total_size(mocker, large: bool = False):
    """Mock the getTotalSize openAPI.

    Arguments:
        mocker: The mocker fixture.
        large: Whether the dataset size is large than free storage.

    Returns:
        The patched mocker and response data.

    """
    response_data = {"totalSize": 7 if not large else sys.maxsize}
    return (
        mocker.patch(
            f"{gas.__name__}.Client.open_api_do",
            return_value=mock_response(data=response_data),
        ),
        response_data,
    )


@function_fixture
def mock_get_users(mocker, is_valid):
    """Mock the getUsers OpenAPI.

    Arguments:
        mocker: The mocker fixture.
        is_valid: Whether the mock need raise UnauthorizedError.

    Returns:
        The patched mocker and response data.

    """
    response_data = {
        "id": "3713a28*************************",
        "nickname": "test",
        "email": "test***@graviti.cn",
        "mobile": "180********",
        "description": "",
        "team": {
            "id": "7d3e****************************",
            "name": "Test",
            "email": None,
            "description": "",
        },
    }
    mock = mocker.patch(
        f"{gas.__name__}.Client.open_api_do", return_value=mock_response(data=response_data)
    )
    if is_valid is False:
        mock.side_effect = UnauthorizedError()

    return mock, response_data



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=import-outside-toplevel

"""Definitions of TensorBay Command-line Interface commands."""

from functools import partial
from typing import Iterable, Optional, Tuple

import click

from tensorbay import __version__
from tensorbay.cli.custom import CustomCommand, DeprecatedOptionsCommand
from tensorbay.cli.utility import ContextInfo


@click.group(context_settings={"help_option_names": ("-h", "--help")})
@click.version_option(__version__)
@click.option(
    "-k",
    "--key",
    "access_key",
    type=str,
    default="",
    help="The AccessKey to TensorBay (replace the AccessKey in config file).",
)
@click.option("-u", "--url", type=str, default="", help="The login url.", hidden=True)
@click.option(
    "-p",
    "--profile",
    "profile_name",
    type=str,
    default="default",
    help="Choose a profile from the config file to login.",
)
@click.option("-d", "--debug", is_flag=True, help="Debug mode.")
@click.pass_context
def cli(ctx: click.Context, access_key: str, url: str, profile_name: str, debug: bool) -> None:
    """You can use ``gas <command>`` to operate on your datasets.\f

    Arguments:
        ctx: The context to be passed as the first argument.
        access_key: The accessKey of gas.
        url: The login URL.
        profile_name: The environment to login.
        debug: Debug mode flag.

    """  # noqa: D301,D415
    from tensorbay.cli.utility import _implement_cli

    _implement_cli(ctx, access_key, url, profile_name, debug)


command = partial(cli.command, cls=CustomCommand)


@command(
    synopsis=(
        "$ gas ls                                       # List all the datasets.",
        "$ gas ls tb:<dataset_name>                     # List segments of a dataset.",
        "$ gas ls tb:<dataset_name>:<segment_name>      # List files of a segment.",
    )
)
@click.argument("tbrn", type=str, default="")
@click.option(
    "-a", "--all", "list_all_files", is_flag=True, help="List all files under the segment."
)
@click.option("-l", "show_total_num", is_flag=True, help="Show the total number of resources")
@click.pass_obj
def ls(  # pylint: disable=invalid-name
    obj: ContextInfo, tbrn: str, list_all_files: bool, show_total_num: bool
) -> None:
    """List data under the path. If path is empty, list the names of all datasets.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: Path to be listed, like "tb:KITTI:seg1". If empty, list names of all datasets.
        list_all_files: If true, list all files under the segment.
        show_total_num: If true, show the total number of resources.

    """  # noqa: D301,D415
    from tensorbay.cli.ls import _implement_ls

    _implement_ls(obj, tbrn, list_all_files, show_total_num)


@command(
    synopsis=(
        "$ gas config [key]                # Show the config.",
        "$ gas config <key> <value>        # Set the config.",
        "$ gas config -u [key]             # Unset the config.",
    )
)
@click.argument("key", type=str, default="")
@click.argument("value", type=str, default="")
@click.option("-u", "--unset", is_flag=True, help="Unset the config option")
@click.pass_obj
def config(obj: ContextInfo, key: str, value: str, unset: bool) -> None:
    """Configure the options when using gas CLI.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        key: The option key.
        value: The option value.
        unset: Whether to unset the option.

    """  # noqa: D301,D415
    from tensorbay.cli.config import _implement_config

    _implement_config(obj, key, value, unset)


@command(
    synopsis=(
        "$ gas dataset                                  # List all the datasets.",
        "$ gas dataset tb:<dataset_name>                # Create a dataset.",
        "$ gas dataset -d [-y] tb:<dataset_name>        # Delete a dataset.",
    )
)
@click.argument("tbrn", type=str, default="")
@click.option("-d", "--delete", "is_delete", is_flag=True, help="Delete TensorBay dataset")
@click.option("-y", "--yes", is_flag=True, help="Confirm to delete the dataset completely.")
@click.pass_obj
def dataset(obj: ContextInfo, tbrn: str, is_delete: bool, yes: bool) -> None:
    """List, create or delete datasets.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The tbrn of the dataset, like "tb:KITTI".
        is_delete: Whether to delete the TensorBay dataset.
        yes: Confirm to delete the dataset completely.

    """  # noqa: D301,D415
    from tensorbay.cli.dataset import _implement_dataset

    _implement_dataset(obj, tbrn, is_delete, yes)


@command(
    cls=DeprecatedOptionsCommand,
    synopsis=(
        "$ gas draft -l tb:<dataset_name>                                # List drafts.",
        "$ gas draft tb:<dataset_name>[@<branch_name>] [-m <title>]      # Create a draft.",
    ),
)
@click.argument("tbrn", type=str)
@click.option("-l", "--list", "is_list", is_flag=True, help="List the drafts.")
@click.option("-e", "--edit", is_flag=True, help="Edit the draft's title and description.")
@click.option("-c", "--close", is_flag=True, help="Close the draft.")
@click.option(
    "-m",
    "--message",
    type=str,
    multiple=True,
    default=(),
    help="The title of the draft.",
)
@click.pass_obj
def draft(  # pylint: disable=too-many-arguments
    obj: ContextInfo,
    tbrn: str,
    is_list: bool,
    edit: bool,
    close: bool,
    message: Tuple[str, ...],
) -> None:
    """List or create drafts.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The tbrn of the dataset.
        is_list: Whether to list the drafts.
        edit: Whether to edit the draft's title and description.
        close: Whether to close the draft.
        message: The message of the draft.

    """  # noqa: D301,D415
    from tensorbay.cli.draft import _implement_draft

    _implement_draft(obj, tbrn, is_list, edit, close, message)


@command(
    synopsis=("$ gas commit tb:<dataset_name>#<draft_number> [-m <title>]      # Commit a draft.",)
)
@click.argument("tbrn", type=str)
@click.option(
    "-m", "--message", type=str, multiple=True, default=(), help="The title of the commit."
)
@click.pass_obj
def commit(obj: ContextInfo, tbrn: str, message: Tuple[str, ...]) -> None:
    """Commit drafts.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The path to commit a draft, like "tb:KITTI#1".
        message: The message of the commit.

    """  # noqa: D301,D415
    from tensorbay.cli.commit import _implement_commit

    _implement_commit(obj, tbrn, message)


@command(
    synopsis=(
        "# Upload a file.",
        "$ gas cp <local_path> tb:<dataset_name>#<draft_number>:<segment_name>[://<remote_path]",
        "",
        "# Upload files.",
        "$ gas cp <local_path1> [local_path2 ...] tb:<dataset_name>#<draft_number>:<segment_name>",
        "",
        "# Upload files in a folder.",
        "$ gas cp -r <local_folder> tb:<dataset_name>#<draft_number>:<segment_name>",
    )
)
@click.argument("local_paths", type=str, nargs=-1)
@click.argument("tbrn", type=str, nargs=1)
@click.option(
    "-r", "--recursive", "is_recursive", is_flag=True, help="Copy directories recursively."
)
@click.option("-j", "--jobs", type=int, default=1, help="The number of threads.")
@click.option(
    "-s",
    "--skip",
    "skip_uploaded_files",
    is_flag=True,
    help="Whether to skip the uploaded files.",
)
@click.pass_obj
def cp(  # pylint: disable=invalid-name, too-many-arguments
    obj: ContextInfo,
    local_paths: Iterable[str],
    tbrn: str,
    is_recursive: bool,
    jobs: int,
    skip_uploaded_files: bool,
) -> None:
    """Copy local data to a remote path.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        local_paths: An iterable of local paths contains data to be uploaded.
        tbrn: The path to save the uploaded data, like "tb:KITTI:seg1".
        is_recursive: Whether copy directories recursively.
        jobs: Number of threads to upload data.
        skip_uploaded_files: Whether skip the uploaded files.

    """  # noqa: D301,D415
    from tensorbay.cli.cp import _implement_cp

    _implement_cp(obj, local_paths, tbrn, is_recursive, jobs, skip_uploaded_files)


@command(
    synopsis=(
        "# Remove a segment.",
        "$ gas rm -r tb:<dataset_name>#<draft_number>:<segment_name>",
        "",
        "# Remove a file.",
        "$ gas rm tb:<dataset_name>#<draft_number>:<segment_name>://<remote_path>",
    )
)
@click.argument("tbrn", type=str)
@click.option(
    "-r",
    "--recursive",
    "is_recursive",
    is_flag=True,
    help="Remove directories recursively.",  # pylint: disable=invalid-name
)
@click.pass_obj
def rm(obj: ContextInfo, tbrn: str, is_recursive: bool) -> None:
    """Remove the remote data.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The path to be removed, like "tb:KITTI#1".
        is_recursive: Whether remove directories recursively.

    """  # noqa: D301,D415
    from tensorbay.cli.rm import _implement_rm

    _implement_rm(obj, tbrn, is_recursive)


@command(
    synopsis=(
        "$ gas branch tb:<dataset_name> [--verbose] [--sort=[-]<key>]    # List branches.",
        "$ gas branch tb:<dataset_name>[@<revision>] <branch_name>       # Create a branch.",
        "$ gas branch -d tb:<dataset_name>@<branch_name>                 # Delete a branch.",
    )
)
@click.argument("tbrn", type=str)
@click.argument("name", type=str, default="")
@click.option("-v", "--verbose", is_flag=True, help="Show short commit id and commit message.")
@click.option("-d", "--delete", "is_delete", is_flag=True, help="Delete the branch")
@click.option(
    "--sort",
    default="name",
    help='Sort based on the key given, which can be "name" or "commit_date". '
    "Prefix - to sort in descending order of the value.",
)
@click.pass_obj
def branch(  # pylint: disable=too-many-arguments
    obj: ContextInfo, tbrn: str, name: str, verbose: bool, is_delete: bool, sort: str
) -> None:
    """List, create or delete branches.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The tbrn of the dataset.
        name: The name of the branch to be created.
        verbose: Whether to show the short commit id and commit message.
        is_delete: Whether to delete the branch.
        sort: The key to sort on.

    """  # noqa: D301,D415
    from tensorbay.cli.branch import _implement_branch

    _implement_branch(obj, tbrn, name, verbose, is_delete, sort)


@command(
    synopsis=(
        "$ gas tag tb:<dataset_name> [--sort=[-]<key>]          # List tags.",
        "$ gas tag tb:<dataset_name>[@<revision>] <tag_name>    # Create a tag.",
        "$ gas tag -d tb:<dataset_name>@<tag_name>              # Delete a tag.",
    )
)
@click.argument("tbrn", type=str)
@click.argument("name", type=str, default="")
@click.option("-d", "--delete", "is_delete", is_flag=True, help="Delete the tag.")
@click.option(
    "--sort",
    default="name",
    help='Sort based on the key given, which can be "name" or "commit_date". '
    "Prefix - to sort in descending order of the value.",
)
@click.pass_obj
def tag(obj: ContextInfo, tbrn: str, name: str, is_delete: bool, sort: str) -> None:
    """List, create or delete tags.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The tbrn of the dataset.
        name: The name of the tag.
        is_delete: Whether to delete the tag.
        sort: The key to sort on.

    """  # noqa: D301,D415
    from tensorbay.cli.tag import _implement_tag

    _implement_tag(obj, tbrn, name, is_delete, sort)


@command(
    synopsis=(
        "# Show commit logs.",
        "$ gas log tb:<dataset_name>[@<revision>]",
        "",
        "# Show up to <number> commit logs.",
        "$ gas log -n <number> tb:<dataset_name>[@<revision>]",
        "",
        "# Show oneline commit logs.",
        "$ gas log --oneline tb:<dataset_name>[@<revision>]",
        "",
        "# Show all commit logs.",
        "$ gas log --all tb:<dataset_name>",
        "",
        "# Show text-based graphical commit logs.",
        "$ gas log --graph tb:<dataset_name>[@<revision>]",
        "# Show commit and open draft logs.",
        "$ gas log --show-drafts tb:<dataset_name>[@<revision>]",
        "",
    )
)
@click.argument("tbrn", type=str)
@click.option(
    "-n", "--max-count", type=int, default=None, help="Limit the max number of commits to show"
)
@click.option("--oneline", is_flag=True, help="Limit commit message to oneline")
@click.option("--all", "is_all", is_flag=True, help="Show all the commits of all branches")
@click.option("--graph", is_flag=True, help="Show text-based graphical commits history")
@click.option("--show-drafts", is_flag=True, help="Show open drafts")
@click.pass_obj
def log(  # pylint: disable=too-many-arguments
    obj: ContextInfo,
    tbrn: str,
    max_count: Optional[int],
    oneline: bool,
    is_all: bool,
    graph: bool,
    show_drafts: bool,
) -> None:
    """Show commit logs.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        tbrn: The tbrn of a dataset.
        max_count: Max number of commits to show.
        oneline: Whether to show a commit message in oneline.
        is_all: Whether to show all commits of all branches.
        graph: Whether to show graphical commit history.
        show_drafts: Whether to log open drafts.

    """  # noqa: D301,D415
    from tensorbay.cli.log import _implement_log

    _implement_log(obj, tbrn, max_count, oneline, is_all, graph, show_drafts)


@command(
    synopsis=(
        "$ gas auth                      # Interactive authentication.",
        "$ gas auth <AccessKey>          # Authenticate with AccessKey.",
        "$ gas auth -g [-a]              # Get the authentication info.",
        "$ gas auth -s [-a]              # Get the authentication and user info.",
        "$ gas auth -u [-a]              # Unset the authentication info.",
    )
)
@click.argument("arg1", type=str, default="", metavar="accessKey")
@click.argument("arg2", type=str, default="", metavar="")
@click.option("-g", "--get", is_flag=True, help="Get the accesskey of the profile")
@click.option(
    "-s", "--status", is_flag=True, help="Get the user info and accesskey of the profile."
)
@click.option("-u", "--unset", is_flag=True, help="Unset the accesskey of the profile")
@click.option("-a", "--all", "is_all", is_flag=True, help="All the auth info")
@click.pass_obj
def auth(  # pylint: disable=too-many-arguments
    obj: ContextInfo,
    arg1: str,
    arg2: str,
    get: bool,
    status: bool,
    unset: bool,
    is_all: bool,
) -> None:
    """Authenticate the accessKey of gas.\f

    Arguments:
        obj: A :class:`.utility.ContextInfo` instance containing the command context.
        arg1: The accessKey or the url of gas for the authentication.
        arg2: The accessKey of gas for the authentication if arg1 is url.
        get: Whether to get the accesskey of the profile.
        status: Whether to get the user info and accesskey of the profile.
        unset: Whether to unset the accesskey of the profile.
        is_all: All the auth info or not.

    """  # noqa: D301,D415
    from tensorbay.cli.auth import _implement_auth

    _implement_auth(obj, arg1, arg2, get, status, unset, is_all)


if __name__ == "__main__":
    cli()  # pylint: disable=no-value-for-parameter



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Client module."""

from tensorbay.client.gas import GAS
from tensorbay.client.profile import profile
from tensorbay.utility import config

__all__ = [
    "GAS",
    "config",
    "profile",
]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Items and AttributeInfo.

:class:`AttributeInfo` represents the information of an attribute.
It refers to the `Json schema`_ method to describe an attribute.

:class:`Items` is the base class of :class:`AttributeInfo`, representing the items of an attribute.

.. _Json schema: https://json-schema.org/

"""

from enum import Enum
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, TypeVar, Union

from tensorbay.utility import EqMixin, NameMixin, ReprMixin, attr, attr_base, camel, common_loads

_AvailaleType = Union[list, bool, int, float, str, None]
_SingleArgType = Union[str, None, Type[_AvailaleType]]
_ArgType = Union[_SingleArgType, Iterable[_SingleArgType]]
_EnumElementType = Union[str, float, bool, None]


class _AttributeType(Enum):
    """All the possible type of the attributes."""

    # pylint: disable=invalid-name
    array = list
    boolean = bool
    integer = int
    number = float
    string = str
    null = None
    instance = "instance"

    @classmethod
    def get_type_name(cls, type_: _SingleArgType) -> str:
        """Return the corresponding enumeration type name of the given string or type.

        Arguments:
            type_: A string or type indicating the attribute type.

        Returns:
            The name of the :class:`_AttributeType` object corresponding to the given type_.

        Raises:
            ValueError: When the input type_ is not supported.

        """
        if type_ in cls.__members__:
            return type_  # type: ignore[return-value]

        try:
            return cls(type_).name
        except ValueError as error:
            raise ValueError(
                "Invalid type_ values for attribute. "
                f"Only support {tuple(cls.__members__.keys())} attribute types"
            ) from error


class Items(ReprMixin, EqMixin):
    """The base class of :class:`AttributeInfo`, representing the items of an attribute.

    When the value type of an attribute is array,
    the :class:`AttributeInfo` would contain an 'items' field.

    .. todo::
        The format of argument *type_* on the generated web page is incorrect.

    Arguments:
        type_: The type of the attribute value, could be a single type or multi-types.
            The type must be within the followings:

                - array
                - boolean
                - integer
                - number
                - string
                - null
                - instance

        enum: All the possible values of an enumeration attribute.
        minimum: The minimum value of number type attribute.
        maximum: The maximum value of number type attribute.
        items: The items inside array type attributes.

    Attributes:
        type: The type of the attribute value, could be a single type or multi-types.
        enum: All the possible values of an enumeration attribute.
        minimum: The minimum value of number type attribute.
        maximum: The maximum value of number type attribute.
        items: The items inside array type attributes.

    Raises:
        TypeError: When both ``enum`` and ``type_`` are absent or
            when ``type_`` is array and ``items`` is absent.

    Examples:
        >>> Items(type_="integer", enum=[1, 2, 3, 4, 5], minimum=1, maximum=5)
        Items(
          (type): 'integer',
          (enum): [...],
          (minimum): 1,
          (maximum): 5
        )

    """

    _T = TypeVar("_T", bound="Items")

    _repr_attrs: Tuple[str, ...] = ("type", "enum", "minimum", "maximum", "items")

    def __init__(
        self,
        *,
        type_: _ArgType = "",
        enum: Optional[Iterable[_EnumElementType]] = None,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        items: Optional["Items"] = None,
    ):
        if type_ != "":
            self.type, has_array = self._convert_type(type_)
            if has_array and items:
                self.items = items

        if enum:
            self.enum = list(enum)
        if minimum is not None:
            self.minimum = minimum
        if minimum is not None:
            self.maximum = maximum

    @staticmethod
    def _convert_type(type_: _ArgType) -> Tuple[Union[str, List[str]], bool]:
        if isinstance(type_, Iterable) and not isinstance(type_, str):
            converted_types = [_AttributeType.get_type_name(single_type) for single_type in type_]
            return converted_types, "array" in converted_types

        converted_type = _AttributeType.get_type_name(type_)
        return converted_type, converted_type == "array"

    def _loads(self, contents: Dict[str, Any]) -> None:
        if "type" in contents:
            self.type, has_array = self._convert_type(contents["type"])
            if has_array:
                self.items = Items.loads(contents["items"])

        if "enum" in contents:
            self.enum = contents["enum"]

        if "minimum" in contents:
            self.minimum = contents["minimum"]

        if "maximum" in contents:
            self.maximum = contents["maximum"]

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Load an Items from a dict containing the items information.

        Arguments:
            contents: A dict containing the information of the items.

        Returns:
            The loaded :class:`Items` object.

        Examples:
            >>> contents = {
            ...     "type": "array",
            ...     "enum": [1, 2, 3, 4, 5],
            ...     "minimum": 1,
            ...     "maximum": 5,
            ...     "items": {
            ...         "enum": [None],
            ...         "type": "null",
            ...     },
            ... }
            >>> Items.loads(contents)
            Items(
              (type): 'array',
              (enum): [...],
              (minimum): 1,
              (maximum): 5,
              (items): Items(...)
            )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the information of the items into a dict.

        Returns:
            A dict containing all the information of the items.

        Examples:
            >>> items = Items(type_="integer", enum=[1, 2, 3, 4, 5], minimum=1, maximum=5)
            >>> items.dumps()
            {'type': 'integer', 'enum': [1, 2, 3, 4, 5], 'minimum': 1, 'maximum': 5}

        """
        contents: Dict[str, Any] = {}
        if hasattr(self, "type"):
            contents["type"] = self.type

        if hasattr(self, "items"):
            contents["items"] = self.items.dumps()

        if hasattr(self, "enum"):
            contents["enum"] = self.enum

        if hasattr(self, "minimum"):
            contents["minimum"] = self.minimum

        if hasattr(self, "maximum"):
            contents["maximum"] = self.maximum

        return contents


class AttributeInfo(NameMixin, Items):
    """This class represents the information of an attribute.

    It refers to the `Json schema`_ method to describe an attribute.

    .. todo::
        The format of argument *type_* on the generated web page is incorrect.

    Arguments:
        name: The name of the attribute.
        type_: The type of the attribute value, could be a single type or multi-types.
            The type must be within the followings:

                - array
                - boolean
                - integer
                - number
                - string
                - null
                - instance

        enum: All the possible values of an enumeration attribute.
        minimum: The minimum value of number type attribute.
        maximum: The maximum value of number type attribute.
        items: The items inside array type attributes.
        parent_categories: The parent categories of the attribute.
        description: The description of the attribute.

    Attributes:
        type: The type of the attribute value, could be a single type or multi-types.
        enum: All the possible values of an enumeration attribute.
        minimum: The minimum value of number type attribute.
        maximum: The maximum value of number type attribute.
        items: The items inside array type attributes.
        parent_categories: The parent categories of the attribute.
        description: The description of the attribute.

    .. _Json schema: https://json-schema.org/

    Examples:
        >>> from tensorbay.label import Items
        >>> items = Items(type_="integer", enum=[1, 2, 3, 4, 5], minimum=1, maximum=5)
        >>> AttributeInfo(
        ...     name="example",
        ...     type_="array",
        ...     enum=[1, 2, 3, 4, 5],
        ...     items=items,
        ...     minimum=1,
        ...     maximum=5,
        ...     parent_categories=["parent_category_of_example"],
        ...     description="This is an example",
        ... )
        AttributeInfo("example")(
          (type): 'array',
          (enum): [
            1,
            2,
            3,
            4,
            5
          ],
          (minimum): 1,
          (maximum): 5,
          (items): Items(
            (type): 'integer',
            (enum): [...],
            (minimum): 1,
            (maximum): 5
          ),
          (parent_categories): [
            'parent_category_of_example'
          ]
        )

    """

    _T = TypeVar("_T", bound="AttributeInfo")

    _repr_attrs = Items._repr_attrs + ("parent_categories",)
    _repr_maxlevel = 2

    _attrs_base: Items = attr_base(key=None)
    parent_categories: List[str] = attr(is_dynamic=True, key=camel)

    def __init__(
        self,
        name: str,
        *,
        type_: _ArgType = "",
        enum: Optional[Iterable[_EnumElementType]] = None,
        minimum: Optional[float] = None,
        maximum: Optional[float] = None,
        items: Optional[Items] = None,
        parent_categories: Union[None, str, Iterable[str]] = None,
        description: str = "",
    ):
        NameMixin.__init__(self, name, description)
        Items.__init__(self, type_=type_, enum=enum, minimum=minimum, maximum=maximum, items=items)

        if not parent_categories:
            return

        if isinstance(parent_categories, str):
            self.parent_categories = [parent_categories]
        else:
            self.parent_categories = list(parent_categories)

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Load an AttributeInfo from a dict containing the attribute information.

        Arguments:
            contents: A dict containing the information of the attribute.

        Returns:
            The loaded :class:`AttributeInfo` object.

        Examples:
            >>> contents = {
            ...     "name": "example",
            ...     "type": "array",
            ...     "items": {"type": "boolean"},
            ...     "description": "This is an example",
            ...     "parentCategories": ["parent_category_of_example"],
            ... }
            >>> AttributeInfo.loads(contents)
            AttributeInfo("example")(
              (type): 'array',
              (items): Items(
                (type): 'boolean',
              ),
              (parent_categories): [
                'parent_category_of_example'
              ]
            )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the information of this attribute into a dict.

        Returns:
            A dict containing all the information of this attribute.

        Examples:
            >>> from tensorbay.label import Items
            >>> items = Items(type_="integer", minimum=1, maximum=5)
            >>> attributeinfo = AttributeInfo(
            ...     name="example",
            ...     type_="array",
            ...     items=items,
            ...     parent_categories=["parent_category_of_example"],
            ...     description="This is an example",
            ... )
            >>> attributeinfo.dumps()
            {
                'name': 'example',
                'description': 'This is an example',
                'type': 'array',
                'items': {'type': 'integer', 'minimum': 1, 'maximum': 5},
                'parentCategories': ['parent_category_of_example'],
            }

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Pipeline and PipelineForIterable.

:class:`Pipeline` runs registered checker functions while healthchecking.
:class:`PipelineForIterable` is a healthcheck pipeline to processes iterable objects.

Checker functions can be registered to the healthcheck pipeline by :meth:`Pipeline.register` or
:meth:`PipelineForIterable.register <Pipeline.register>`.

"""

from typing import Callable, Generic, Iterable, Iterator, List, TypeVar

_A = TypeVar("_A")
_R = TypeVar("_R")
_Checker = Callable[[_A], Iterator[_R]]


class Pipeline(Generic[_A, _R]):
    """:class:`Pipeline` is a healthcheck pipeline to run registered checker functions."""

    _S = TypeVar("_S", bound="Pipeline[_A, _R]")

    def __init__(self) -> None:
        self._pipeline: List[_Checker[_A, _R]] = []

    def __call__(self, args: _A) -> Iterator[_R]:
        """Call the :class:`Pipeline` as a function.

        Arguments:
            args: The information needs to be checked.

        Yields:
            The checker function.

        """
        for checker in self._pipeline:
            yield from checker(args)

    def register(self, checker: _Checker[_A, _R]) -> _Checker[_A, _R]:
        """Decorator function to register checkers into pipeline.

        Arguments:
            checker: The checker function needs to be registered.

        Returns:
            The checker function unchanged.

        """
        self._pipeline.append(checker)
        return checker

    def copy(self: _S) -> _S:
        """Copy method to get a shallow copy of pipeline.

        Returns:
            A shallow copy of the pipeline.

        """
        pipeline = self.__class__()
        pipeline._pipeline = self._pipeline.copy()  # pylint: disable=protected-access
        return pipeline


class PipelineForIterable(Pipeline[_A, _R]):
    """Healthcheck pipeline for processing iterable objects."""

    def __call__(self, args: Iterable[_A]) -> Iterator[_R]:  # type: ignore[override]
        """Call the :class:`Pipeline` as a function.

        Arguments:
            args: The information needs to be checked.

        Yields:
            The checker function.

        """
        for arg in args:
            yield from super().__call__(arg)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Interact with sextant app at graviti marketplace."""

import time
from typing import Any, Dict, Generator, List, Optional
from urllib.parse import urljoin

from tensorbay.client.lazy import PagingList
from tensorbay.client.requests import Client
from tensorbay.exception import ResourceNotExistError


class Evaluation:
    """This class defines :class:`Evaluation`.

    Arguments:
        evaluation_id: Evaluation ID.
        created_at: Created time of the evaluation.
        benchmark: The :class:`Benchmark`.

    """

    def __init__(self, evaluation_id: str, created_at: int, benchmark: "Benchmark") -> None:
        self.evaluation_id = evaluation_id
        self.benchmark = benchmark
        self.created_at = created_at

    def __repr__(self) -> str:
        read_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.created_at))
        return f"{self.__class__.__name__} createdAt: {read_time}"

    def get_result(self) -> Dict[str, Any]:
        """Get the result of the evaluation.

        Returns:
            The result dict of the evaluation.

        """
        return self.benchmark.sextant.open_api_do(  # type: ignore[no-any-return]
            "GET",
            f"benchmarks/{self.benchmark.benchmark_id}/evaluations/{self.evaluation_id}/results",
            "",
        ).json()

    # def get_status(self) -> str:
    #     """Get the status of the evaluation.

    #     Return:
    #         One of "success", "fail" or "processing".

    #     """


class Benchmark:  # pylint: disable=too-many-instance-attributes
    """This class defines :class:`Benchmark`.

    Arguments:
        name: Name of the Benchmark.
        dataset_id: ID of the dataset on which this benchmark based.
        commit_id: ID of the commit which is used as the evaluation benchmark.
        benchmark_id: Benchmark ID.
        sextant: The :class:`SextantClient`.
        categories: The needed evaluation categories, if not given, all categories will be used.
        iou_threshold: The IoU threshold.
        customized_metrics: Https url of the github repository.

    """

    def __init__(
        self,
        name: str,
        benchmark_id: str,
        sextant: "Sextant",
        *,
        dataset_id: Optional[str] = None,
        commit_id: Optional[str] = None,
        categories: Optional[List[str]] = None,
        iou_threshold: Optional[float] = None,
        customized_metrics: Optional[str] = None,
    ) -> None:
        self.name = name
        self.dataset_id = dataset_id
        self.benchmark_id = benchmark_id
        self.commit_id = commit_id
        self.sextant = sextant
        self.categories = categories
        self.iou_threshold = iou_threshold
        self.customized_metrics = customized_metrics

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    def _generate_evaluations(
        self, offset: int = 0, limit: int = 128
    ) -> Generator[Evaluation, None, int]:

        params: Dict[str, Any] = {"offset": offset, "limit": limit}
        response = self.sextant.open_api_do(
            "GET", f"benchmarks/{self.benchmark_id}/evaluations", "", params=params
        ).json()

        for evaluation in response["evaluations"]:
            yield Evaluation(evaluation["evaluationId"], evaluation["createdAt"], self)

        return response["totalCount"]  # type: ignore[no-any-return]

    def create_evaluation(self, dataset_id: str, commit_id: str) -> Evaluation:
        """Create an evaluation task.

        Arguments:
            dataset_id: Id of the needed evaluation dataset.
            commit_id: Id of the needed commit.

        Returns:
            The created evaluation instance.

        """
        post_data = {"datasetId": dataset_id, "commitId": commit_id}
        evaluation_id = self.sextant.open_api_do(
            "POST", f"benchmarks/{self.benchmark_id}/evaluations", "", json=post_data
        ).json()["evaluationId"]
        return Evaluation(evaluation_id, int(time.time()), self)

    def list_evaluations(self) -> PagingList[Evaluation]:
        """List all evaluations.

        Returns:
            A list of evaluations.

        """
        return PagingList(self._generate_evaluations, 128)


class Sextant(Client):
    """This class defines :class:`Sextant`.

    Arguments:
        access_key: User's access key.
        url: The URL of the graviti gas website.

    """

    def __init__(self, access_key: str, url: str = "") -> None:
        super().__init__(access_key, url)
        self._open_api = urljoin(self.gateway_url, "apps-sextant/v1/")

    def _generate_benmarks(
        self, offset: int = 0, limit: int = 128
    ) -> Generator[Benchmark, None, int]:

        params: Dict[str, Any] = {"offset": offset, "limit": limit}
        response = self.open_api_do("GET", "benchmarks", "", params=params).json()

        for benchmark in response["benchmarks"]:
            yield Benchmark(benchmark["name"], benchmark["benchmarkId"], self)

        return response["totalCount"]  # type: ignore[no-any-return]

    def create_benchmark(
        self,
        name: str,
        dataset_id: str,
        commit_id: str,
        *,
        categories: Optional[List[str]] = None,
        iou_threshold: Optional[float] = None,
        customized_metrics: Optional[str] = None,
    ) -> Benchmark:
        """Create a benchmark with the given parameters.

        Arguments:
            name: Name of the Benchmark.
            dataset_id: ID of the dataset on which this benchmark based.
            commit_id: ID of the commit which used as the evaluation benchmark.
            categories: The needed evaluation categories, if not given, all categories will be used.
            iou_threshold: The IoU threshold.
            customized_metrics: Https url of the github repository.

        Raises:# flake8: noqa: F402
            ValueError: When iou_threshold and customized_metrics both given or both not given.

        """

    def list_benchmarks(self) -> PagingList[Benchmark]:
        """List all benchmarks.

        Returns:
            The list of Benchmark instances.

        """
        return PagingList(self._generate_benmarks, 128)

    def get_benchmark(self, name: str) -> Benchmark:
        """Get a benchmark instance by name.

        Arguments:
            name: Name of the Benchmark.

        Returns:
            The Benchmark instance with the given name.

        Raises:
            ResourceNotExistError: When the required benchmark does not exist.

        """
        for benchmark in self.list_benchmarks():
            if benchmark.name == name:
                return benchmark

        raise ResourceNotExistError(resource="benchmark", identification=name)



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""SensorType, Sensor, Lidar, Radar, Camera, FisheyeCamera and Sensors.

:class:`SensorType` is an enumeration type. It includes 'LIDAR', 'RADAR', 'CAMERA' and
'FISHEYE_CAMERA'.

:class:`Sensor` defines the concept of sensor. It includes name, description, translation
and rotation.

A :class:`Sensor` class can be initialized by :meth:`Sensor.__init__()` or
:meth:`Sensor.loads()` method.

:class:`Lidar` defines the concept of lidar. It is a kind of sensor for measuring distances by
illuminating the target with laser light and measuring the reflection.

:class:`Radar` defines the concept of radar. It is a detection system that uses radio waves to
determine the range, angle, or velocity of objects.

:class:`Camera` defines the concept of camera. It includes name, description, translation,
rotation, cameraMatrix and distortionCoefficients.

:class:`FisheyeCamera` defines the concept of fisheye camera. It is an ultra wide-angle lens that
produces strong visual distortion intended to create a wide panoramic or hemispherical image.

:class:`Sensors` represent all the sensors in a :class:`~tensorbay.dataset.segment.FusionSegment`.

"""

import warnings
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, TypeVar, Union

from tensorbay.geometry import Transform3D
from tensorbay.sensor.intrinsics import CameraIntrinsics
from tensorbay.utility import (
    MatrixType,
    NameMixin,
    ReprType,
    SortedNameList,
    TypeEnum,
    TypeMixin,
    TypeRegister,
    common_loads,
)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from quaternion import quaternion as Quaternion

_T = TypeVar("_T", bound="Sensor")


class SensorType(TypeEnum):
    """SensorType is an enumeration type.

    It includes 'LIDAR', 'RADAR', 'CAMERA' and 'FISHEYE_CAMERA'.

    Examples:
        >>> SensorType.CAMERA
        <SensorType.CAMERA: 'CAMERA'>
        >>> SensorType["CAMERA"]
        <SensorType.CAMERA: 'CAMERA'>

        >>> SensorType.CAMERA.name
        'CAMERA'
        >>> SensorType.CAMERA.value
        'CAMERA'

    """

    LIDAR = "LIDAR"
    RADAR = "RADAR"
    CAMERA = "CAMERA"
    FISHEYE_CAMERA = "FISHEYE_CAMERA"


class Sensor(NameMixin, TypeMixin[SensorType]):
    """Sensor defines the concept of sensor.

    :class:`Sensor` includes name, description, translation and rotation.

    Arguments:
        name: :class:`Sensor`'s name.

    Raises:
        TypeError: Can not instantiate abstract class :class:`Sensor`.

    Attributes:
        extrinsics: The translation and rotation of the sensor.

    """

    _Type = Union["Radar", "Lidar", "FisheyeCamera", "Camera"]

    _repr_type = ReprType.INSTANCE
    _repr_attrs: Tuple[str, ...] = ("extrinsics",)
    _repr_maxlevel = 3

    extrinsics: Transform3D

    def __new__(cls: Type[_T], name: str) -> _T:  # pylint: disable=unused-argument,arguments-differ
        """Create a new instance of Sensor.

        Arguments:
            name: :class:`Sensor`'s name.

        Returns:
            A :class:`Sensor` instance containing the :class:`sensor`'s name.

        Raises:
            TypeError: Can't instantiate abstract class :class:`Sensor`.

        """
        if cls is Sensor:
            raise TypeError("Can't instantiate abstract class Sensor")

        obj: _T = object.__new__(cls)
        return obj

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def _loads(self, contents: Dict[str, Any]) -> None:
        super()._loads(contents)
        extrinsics = contents.get("extrinsics")
        if extrinsics:
            self.extrinsics = Transform3D.loads(contents["extrinsics"])

    @staticmethod
    def loads(contents: Dict[str, Any]) -> "_Type":
        """Loads a Sensor from a dict containing the sensor information.

        Arguments:
            contents: A dict containing name, description and sensor extrinsics.

        Returns:
            A :class:`Sensor` instance containing the information from the contents dict.

        Examples:
            >>> contents = {
            ...     "name": "Lidar1",
            ...     "type": "LIDAR",
            ...     "extrinsics": {
            ...         "translation": {"x": 1.1, "y": 2.2, "z": 3.3},
            ...         "rotation": {"w": 1.1, "x": 2.2, "y": 3.3, "z": 4.4},
            ...     },
            ... }
            >>> sensor = Sensor.loads(contents)
            >>> sensor
            Lidar("Lidar1")(
                (extrinsics): Transform3D(
                    (translation): Vector3D(1.1, 2.2, 3.3),
                    (rotation): quaternion(1.1, 2.2, 3.3, 4.4)
                )
            )

        """
        sensor: "Sensor._Type" = common_loads(SensorType(contents["type"]).type, contents)
        return sensor

    def dumps(self) -> Dict[str, Any]:
        """Dumps the sensor into a dict.

        Returns:
            A dict containing the information of the sensor.

        Examples:
            >>> # sensor is the object initialized from self.loads() method.
            >>> sensor.dumps()
            {
                'name': 'Lidar1',
                'type': 'LIDAR',
                'extrinsics': {'translation': {'x': 1.1, 'y': 2.2, 'z': 3.3},
                'rotation': {'w': 1.1, 'x': 2.2, 'y': 3.3, 'z': 4.4}
                }
            }

        """
        contents: Dict[str, Any] = super()._dumps()
        contents["type"] = self.enum.value
        if hasattr(self, "extrinsics"):
            contents["extrinsics"] = self.extrinsics.dumps()

        return contents

    def set_extrinsics(
        self,
        translation: Iterable[float] = (0, 0, 0),
        rotation: Transform3D.RotationType = (1, 0, 0, 0),
        *,
        matrix: Optional[MatrixType] = None,
    ) -> None:
        """Set the extrinsics of the sensor.

        Arguments:
            translation: Translation parameters.
            rotation: Rotation in a sequence of [w, x, y, z] or numpy quaternion.
            matrix: A 3x4 or 4x4 transform matrix.

        Examples:
            >>> sensor.set_extrinsics(translation=translation, rotation=rotation)
            >>> sensor
            Lidar("Lidar1")(
                (extrinsics): Transform3D(
                    (translation): Vector3D(1, 2, 3),
                    (rotation): quaternion(1, 2, 3, 4)
                )
            )

        """
        self.extrinsics = Transform3D(translation, rotation, matrix=matrix)

    def set_translation(self, x: float, y: float, z: float) -> None:
        """Set the translation of the sensor.

        Arguments:
            x: The x coordinate of the translation.
            y: The y coordinate of the translation.
            z: The z coordinate of the translation.

        Examples:
            >>> sensor.set_translation(x=2, y=3, z=4)
            >>> sensor
            Lidar("Lidar1")(
                (extrinsics): Transform3D(
                    (translation): Vector3D(2, 3, 4),
                    ...
                )
            )

        """
        if not hasattr(self, "extrinsics"):
            self.extrinsics = Transform3D()
        self.extrinsics.set_translation(x, y, z)

    def set_rotation(
        self,
        w: Optional[float] = None,
        x: Optional[float] = None,
        y: Optional[float] = None,
        z: Optional[float] = None,
        *,
        quaternion: Optional[Quaternion] = None,
    ) -> None:
        """Set the rotation of the sensor.

        Arguments:
            w: The w componet of the roation quaternion.
            x: The x componet of the roation quaternion.
            y: The y componet of the roation quaternion.
            z: The z componet of the roation quaternion.
            quaternion: Numpy quaternion representing the rotation.

        Examples:
            >>> sensor.set_rotation(2, 3, 4, 5)
            >>> sensor
            Lidar("Lidar1")(
                (extrinsics): Transform3D(
                    ...
                    (rotation): quaternion(2, 3, 4, 5)
                )
            )

        """
        if not hasattr(self, "extrinsics"):
            self.extrinsics = Transform3D()
        self.extrinsics.set_rotation(w, x, y, z, quaternion=quaternion)


@TypeRegister(SensorType.LIDAR)
class Lidar(Sensor):
    """Lidar defines the concept of lidar.

    :class:`Lidar` is a kind of sensor for measuring distances by illuminating the target
    with laser light and measuring the reflection.

    Examples:
        >>> lidar = Lidar("Lidar1")
        >>> lidar.set_extrinsics(translation=translation, rotation=rotation)
        >>> lidar
        Lidar("Lidar1")(
            (extrinsics): Transform3D(
                (translation): Vector3D(1, 2, 3),
                (rotation): quaternion(1, 2, 3, 4)
            )
        )

    """


@TypeRegister(SensorType.RADAR)
class Radar(Sensor):
    """Radar defines the concept of radar.

    :class:`Radar` is a detection system that uses radio waves to determine the range, angle,
    or velocity of objects.

    Examples:
        >>> radar = Radar("Radar1")
        >>> radar.set_extrinsics(translation=translation, rotation=rotation)
        >>> radar
        Radar("Radar1")(
            (extrinsics): Transform3D(
                (translation): Vector3D(1, 2, 3),
                (rotation): quaternion(1, 2, 3, 4)
            )
        )

    """


@TypeRegister(SensorType.CAMERA)
class Camera(Sensor):
    """Camera defines the concept of camera.

    :class:`Camera` includes name, description, translation, rotation, cameraMatrix
    and distortionCoefficients.

    Attributes:
        extrinsics: The translation and rotation of the camera.
        intrinsics: The camera matrix and distortion coefficients of the camera.

    Examples:
        >>> from tensorbay.geometry import Vector3D
        >>> from numpy import quaternion
        >>> camera = Camera('Camera1')
        >>> translation = Vector3D(1, 2, 3)
        >>> rotation = quaternion(1, 2, 3, 4)
        >>> camera.set_extrinsics(translation=translation, rotation=rotation)
        >>> camera.set_camera_matrix(fx=1.1, fy=1.1, cx=1.1, cy=1.1)
        >>> camera.set_distortion_coefficients(p1=1.2, p2=1.2, k1=1.2, k2=1.2)
        >>> camera
        Camera("Camera1")(
            (extrinsics): Transform3D(
                (translation): Vector3D(1, 2, 3),
                (rotation): quaternion(1, 2, 3, 4)
            ),
            (intrinsics): CameraIntrinsics(
                (camera_matrix): CameraMatrix(
                    (fx): 1.1,
                    (fy): 1.1,
                    (cx): 1.1,
                    (cy): 1.1,
                    (skew): 0
                ),
                (distortion_coefficients): DistortionCoefficients(
                    (p1): 1.2,
                    (p2): 1.2,
                    (k1): 1.2,
                    (k2): 1.2
                )
            )
        )

    """

    _T = TypeVar("_T", bound="Camera")

    _repr_attrs = Sensor._repr_attrs + ("intrinsics",)

    intrinsics: CameraIntrinsics

    def _loads(self, contents: Dict[str, Any]) -> None:
        super()._loads(contents)
        intrinsics = contents.get("intrinsics")
        if intrinsics:
            self.intrinsics = CameraIntrinsics.loads(contents["intrinsics"])

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a Camera from a dict containing the camera information.

        Arguments:
            contents: A dict containing name, description, extrinsics and intrinsics.

        Returns:
            A :class:`Camera` instance containing information from contents dict.

        Examples:
            >>> contents = {
            ...     "name": "Camera1",
            ...     "type": "CAMERA",
            ...     "extrinsics": {
            ...           "translation": {"x": 1, "y": 2, "z": 3},
            ...           "rotation": {"w": 1.0, "x": 2.0, "y": 3.0, "z": 4.0},
            ...     },
            ...     "intrinsics": {
            ...         "cameraMatrix": {"fx": 1, "fy": 1, "cx": 1, "cy": 1, "skew": 0},
            ...         "distortionCoefficients": {"p1": 1, "p2": 1, "k1": 1, "k2": 1},
            ...     },
            ... }
            >>> Camera.loads(contents)
            Camera("Camera1")(
                    (extrinsics): Transform3D(
                        (translation): Vector3D(1, 2, 3),
                        (rotation): quaternion(1, 2, 3, 4)
                    ),
                    (intrinsics): CameraIntrinsics(
                        (camera_matrix): CameraMatrix(
                            (fx): 1,
                            (fy): 1,
                            (cx): 1,
                            (cy): 1,
                            (skew): 0
                        ),
                        (distortion_coefficients): DistortionCoefficients(
                            (p1): 1,
                            (p2): 1,
                            (k1): 1,
                            (k2): 1
                        )
                    )
                )

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps the camera into a dict.

        Returns:
            A dict containing name, description, extrinsics and intrinsics.

        Examples:
            >>> camera.dumps()
            {
                'name': 'Camera1',
                'type': 'CAMERA',
                'extrinsics': {
                    'translation': {'x': 1, 'y': 2, 'z': 3},
                    'rotation': {'w': 1.0, 'x': 2.0, 'y': 3.0, 'z': 4.0}
                },
                'intrinsics': {
                    'cameraMatrix': {'fx': 1, 'fy': 1, 'cx': 1, 'cy': 1, 'skew': 0},
                    'distortionCoefficients': {'p1': 1, 'p2': 1, 'k1': 1, 'k2': 1}
                }
            }

        """
        contents = super().dumps()
        if hasattr(self, "intrinsics"):
            contents["intrinsics"] = self.intrinsics.dumps()
        return contents

    def set_camera_matrix(  # pylint: disable=[too-many-arguments, invalid-name]
        self,
        fx: Optional[float] = None,
        fy: Optional[float] = None,
        cx: Optional[float] = None,
        cy: Optional[float] = None,
        skew: float = 0,
        *,
        matrix: Optional[MatrixType] = None,
    ) -> None:
        """Set camera matrix.

        Arguments:
            fx: The x axis focal length expressed in pixels.
            fy: The y axis focal length expressed in pixels.
            cx: The x coordinate of the so called principal point that should be in the center of
                the image.
            cy: The y coordinate of the so called principal point that should be in the center of
                the image.
            skew: It causes shear distortion in the projected image.
            matrix: Camera matrix in 3x3 sequence.

        Examples:
            >>> camera.set_camera_matrix(fx=1.1, fy=2.2, cx=3.3, cy=4.4)
            >>> camera
            Camera("Camera1")(
                ...
                (intrinsics): CameraIntrinsics(
                    (camera_matrix): CameraMatrix(
                        (fx): 1.1,
                        (fy): 2.2,
                        (cx): 3.3,
                        (cy): 4.4,
                        (skew): 0
                    ),
                    ...
                    )
                )
            )

        """
        if not hasattr(self, "intrinsics"):
            self.intrinsics = CameraIntrinsics(fx, fy, cx, cy, skew, camera_matrix=matrix)
            return

        self.intrinsics.set_camera_matrix(fx, fy, cx, cy, skew, matrix=matrix)

    def set_distortion_coefficients(self, **kwargs: float) -> None:
        """Set distortion coefficients.

        Arguments:
            **kwargs: Float values to set distortion coefficients.

        Raises:
            ValueError: When intrinsics is not set yet.

        Examples:
            >>> camera.set_distortion_coefficients(p1=1.1, p2=2.2, k1=3.3, k2=4.4)
            >>> camera
            Camera("Camera1")(
                ...
                (intrinsics): CameraIntrinsics(
                    ...
                    (distortion_coefficients): DistortionCoefficients(
                        (p1): 1.1,
                        (p2): 2.2,
                        (k1): 3.3,
                        (k2): 4.4
                    )
                )
            )

        """
        if not hasattr(self, "intrinsics"):
            raise ValueError("Camera matrix of camera intrinsics must be set first.")
        self.intrinsics.set_distortion_coefficients(**kwargs)


@TypeRegister(SensorType.FISHEYE_CAMERA)
class FisheyeCamera(Camera):
    """FisheyeCamera defines the concept of fisheye camera.

    Fisheye camera is an ultra wide-angle lens that produces strong visual distortion intended
    to create a wide panoramic or hemispherical image.

    Examples:
        >>> fisheye_camera = FisheyeCamera("FisheyeCamera1")
        >>> fisheye_camera.set_extrinsics(translation=translation, rotation=rotation)
        >>> fisheye_camera
        FisheyeCamera("FisheyeCamera1")(
            (extrinsics): Transform3D(
                (translation): Vector3D(1, 2, 3),
                (rotation): quaternion(1, 2, 3, 4)
            )
        )

    """


class Sensors(SortedNameList[Sensor._Type]):  # pylint: disable=protected-access
    """This class represents all sensors in a :class:`~tensorbay.dataset.segment.FusionSegment`."""

    _T = TypeVar("_T", bound="Sensors")

    def _loads(self, contents: List[Dict[str, Any]]) -> None:
        self._data = []
        self._names = []
        for sensor_info in contents:
            self.add(Sensor.loads(sensor_info))

    @classmethod
    def loads(cls: Type[_T], contents: List[Dict[str, Any]]) -> _T:
        """Loads a :class:`Sensors` instance from the given contents.

        Arguments:
            contents: A list of dict containing the sensors information in a fusion segment,
                whose format should be like::

                    [
                        {
                            "name": <str>
                            "type": <str>
                            "extrinsics": {
                                "translation": {
                                    "x": <float>
                                    "y": <float>
                                    "z": <float>
                                },
                                "rotation": {
                                    "w": <float>
                                    "x": <float>
                                    "y": <float>
                                    "z": <float>
                                },
                            },
                            "intrinsics": {           --- only for cameras
                                "cameraMatrix": {
                                    "fx": <float>
                                    "fy": <float>
                                    "cx": <float>
                                    "cy": <float>
                                    "skew": <float>
                                }
                                "distortionCoefficients": {
                                    "k1": <float>
                                    "k2": <float>
                                    "p1": <float>
                                    "p2": <float>
                                    ...
                                }
                            },
                            "desctiption": <str>
                        },
                        ...
                    ]

        Returns:
            The loaded :class:`Sensors` instance.

        """
        return common_loads(cls, contents)

    def dumps(self) -> List[Dict[str, Any]]:
        """Return the information of all the sensors.

        Returns:
            A list of dict containing the information of all sensors::

                [
                    {
                        "name": <str>
                        "type": <str>
                        "extrinsics": {
                            "translation": {
                                "x": <float>
                                "y": <float>
                                "z": <float>
                            },
                            "rotation": {
                                "w": <float>
                                "x": <float>
                                "y": <float>
                                "z": <float>
                            },
                        },
                        "intrinsics": {           --- only for cameras
                            "cameraMatrix": {
                                "fx": <float>
                                "fy": <float>
                                "cx": <float>
                                "cy": <float>
                                "skew": <float>
                            }
                            "distortionCoefficients": {
                                "k1": <float>
                                "k2": <float>
                                "p1": <float>
                                "p2": <float>
                                ...
                            }
                        },
                        "desctiption": <str>
                    },
                    ...
                ]

        """
        return [sensor.dumps() for sensor in self._data]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of COVID-chestxray dataset."""

import csv
import os
from typing import Any, Dict

from tensorbay.dataset import Data, Dataset
from tensorbay.label import Classification

DATASET_NAME = "COVID-chestxray"

_INTEGER_GROUP = {"offset", "age"}
_FLOAT_GROUP = {
    "temperature",
    "pO2_saturation",
    "leukocyte_count",
    "neutrophil_count",
    "lymphocyte_count",
}


def COVIDChestXRay(path: str) -> Dataset:
    """`COVID-chestxray <https://github.com/ieee8023/covid-chestxray-dataset>`_ dataset.

    The file structure should be like::

        <path>
            images/
                0a7faa2a.jpg
                000001-2.png
                000001-3.jpg
                1B734A89-A1BF-49A8-A1D3-66FAFA4FAC5D.jpeg
                ...
            volumes/
                coronacases_org_001.nii.gz
                ....
            metadata.csv
            ...

    Arguments:
        path: The root directory of the dataset.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    segment = dataset.create_segment()

    csv_path = os.path.join(root_path, "metadata.csv")

    with open(csv_path, encoding="utf-8") as fp:
        csv_reader = csv.DictReader(fp)
        for attributes in csv_reader:
            folder = attributes.pop("folder")
            # The 20 images invovled in "volumes" folder currently are invalid to download.
            if folder == "volumes":
                continue
            image_path = os.path.join(root_path, folder, attributes.pop("filename"))
            category = attributes.pop("finding").strip()
            data = Data(image_path)
            data.label.classification = Classification(
                category=category, attributes=_convert_type(attributes)
            )
            segment.append(data)
    return dataset


def _convert_type(attributes: Dict[str, Any]) -> Dict[str, Any]:
    for key, value in attributes.items():
        if value == "":
            attributes[key] = None
        elif key in _INTEGER_GROUP:
            attributes[key] = int(value)
        elif key in _FLOAT_GROUP:
            attributes[key] = float(value)

    return attributes



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of BSTLD dataset."""

import os

from tensorbay.dataset import Data, Dataset
from tensorbay.exception import ModuleImportError
from tensorbay.label import LabeledBox2D

DATASET_NAME = "BSTLD"

_LABEL_FILENAME_DICT = {
    "test": "test.yaml",
    "train": "train.yaml",
    "additional": "additional_train.yaml",
}


def BSTLD(path: str) -> Dataset:
    """`BSTLD <https://hci.iwr.uni-heidelberg.de/content\
    /bosch-small-traffic-lights-dataset>`_ dataset.

    The file structure should be like::

        <path>
            rgb/
                additional/
                    2015-10-05-10-52-01_bag/
                        <image_name>.jpg
                        ...
                    ...
                test/
                    <image_name>.jpg
                    ...
                train/
                    2015-05-29-15-29-39_arastradero_traffic_light_loop_bag/
                        <image_name>.jpg
                        ...
                    ...
            test.yaml
            train.yaml
            additional_train.yaml

    Arguments:
        path: The root directory of the dataset.

    Raises:
        ModuleImportError: When the module "yaml" can not be found.

    Returns:
        Loaded :class:`~tensorbay.dataset.dataset.Dataset` instance.

    """
    try:
        import yaml  # pylint: disable=import-outside-toplevel
    except ModuleNotFoundError as error:
        raise ModuleImportError(module_name=error.name, package_name="pyyaml") from error

    root_path = os.path.abspath(os.path.expanduser(path))

    dataset = Dataset(DATASET_NAME)
    dataset.load_catalog(os.path.join(os.path.dirname(__file__), "catalog.json"))

    for mode, label_file_name in _LABEL_FILENAME_DICT.items():
        segment = dataset.create_segment(mode)
        label_file_path = os.path.join(root_path, label_file_name)

        with open(label_file_path, encoding="utf-8") as fp:
            labels = yaml.load(fp, yaml.FullLoader)

        for label in labels:
            if mode == "test":
                # the path in test label file looks like:
                # /absolute/path/to/<image_name>.png
                file_path = os.path.join(root_path, "rgb", "test", label["path"].rsplit("/", 1)[-1])
            else:
                # the path in label file looks like:
                # ./rgb/additional/2015-10-05-10-52-01_bag/<image_name>.png
                file_path = os.path.join(root_path, *label["path"][2:].split("/"))
            data = Data(file_path)
            data.label.box2d = [
                LabeledBox2D(
                    box["x_min"],
                    box["y_min"],
                    box["x_max"],
                    box["y_max"],
                    category=box["label"],
                    attributes={"occluded": box["occluded"]},
                )
                for box in label["boxes"]
            ]
            segment.append(data)

    return dataset



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the Animals with attributes 2 dataset."""

from tensorbay.opendataset.AnimalsWithAttributes2.loader import AnimalsWithAttributes2

__all__ = ["AnimalsWithAttributes2"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of VOC2012Detection."""

from tensorbay.opendataset.VOC2012Detection.loader import VOC2012Detection

__all__ = ["VOC2012Detection"]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""User, Commit, Tag, Branch and Draft classes.

:class:`User` defines the basic concept of a user with an action.

:class:`Commit` defines the structure of a commit.

:class:`Tag` defines the structure of a commit tag.

:class:`Branch` defines the structure of a branch.

:class:`Draft` defines the structure of a draft.

"""

from functools import partial
from typing import Any, Dict, Optional, Tuple, Type, TypeVar

from tensorbay.utility import AttrsMixin, NameMixin, ReprMixin, attr, camel, common_loads

ROOT_COMMIT_ID = "00000000000000000000000000000000"


class TeamInfo(NameMixin):
    """This class defines the basic concept of a TensorBay team.

    Arguments:
        name: The name of the team.
        email: The email of the team.
        description: The description of the team.

    """

    _T = TypeVar("_T", bound="TeamInfo")

    _repr_attrs = ("email",)

    email: Optional[str] = attr(default=None)

    def __init__(self, name: str, *, email: Optional[str] = None, description: str = "") -> None:
        super().__init__(name, description)
        self.email = email

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`TeamInfo` instance from the given contents.

        Arguments:
            contents: A dict containing all the information of the commit::

                    {
                        "name": <str>
                        "email": <str>
                        "description": <str>
                    }

        Returns:
            A :class:`TeamInfo` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the information into a dict.

        Returns:
            A dict containing all the information of the team::

                {
                        "name": <str>
                        "email": <str>
                        "description": <str>
                }

        """
        return self._dumps()


class UserInfo(NameMixin):
    """This class defines the basic concept of a TensorBay user.

    Arguments:
        name: The nickname of the user.
        email: The email of the user.
        mobile: The mobile of the user.
        description: The description of the user.
        team: The team of the user.

    """

    _T = TypeVar("_T", bound="UserInfo")

    _repr_attrs = ("email", "mobile", "team")

    _name: str = attr(key="nickname")
    email: Optional[str] = attr(default=None)
    mobile: Optional[str] = attr(default=None)
    team: Optional[TeamInfo] = attr(default=None)

    def __init__(
        self,
        name: str,
        *,
        email: Optional[str] = None,
        mobile: Optional[str] = None,
        description: str = "",
        team: Optional[TeamInfo] = None,
    ) -> None:
        super().__init__(name, description=description)
        self.email = email
        self.mobile = mobile
        self.team = team

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`UserInfo` instance from the given contents.

        Arguments:
            contents: A dict containing all the information of the commit::

                    {
                        "name": <str>
                        "email": <str>
                        "mobile": <str>
                        "description": <str>
                        "team": {  <dict>
                            "name": <str>
                            "email": <str>
                            "description": <str>
                        }
                    }

        Returns:
            A :class:`UserInfo` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the information into a dict.

        Returns:
            A dict containing all the information of the user::

                {
                        "name": <str>
                        "email": <str>
                        "mobile": <str>
                        "description": <str>
                        "team": {  <dict>
                            "name": <str>
                            "email": <str>
                            "description": <str>
                        }
                }

        """
        return self._dumps()


class User(AttrsMixin, ReprMixin):
    """This class defines the basic concept of a user with an action.

    Arguments:
        name: The name of the user.
        date: The date of the user action.

    """

    _T = TypeVar("_T", bound="User")

    _repr_attrs = ("date",)

    name: str = attr()
    date: int = attr()

    def __init__(self, name: str, date: int) -> None:
        self.name = name
        self.date = date

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`User` instance from the given contents.

        Arguments:
            contents: A dict containing all the information of the commit::

                    {
                        "name": <str>
                        "date": <int>
                    }

        Returns:
            A :class:`User` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the user information into a dict.

        Returns:
            A dict containing all the information of the user::

                {
                    "name": <str>
                    "date": <int>
                }

        """
        return self._dumps()


class Commit(AttrsMixin, ReprMixin):
    """This class defines the structure of a commit.

    Arguments:
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.

    """

    _T = TypeVar("_T", bound="Commit")

    _repr_attrs: Tuple[str, ...] = ("parent_commit_id", "title", "committer")
    _repr_maxlevel = 2

    commit_id: str = attr(key=camel)
    parent_commit_id: Optional[str] = attr(key=camel)
    title: str = attr()
    description: str = attr(default="")
    committer: User = attr()

    def __init__(  # pylint: disable=too-many-arguments
        self,
        commit_id: str,
        parent_commit_id: Optional[str],
        title: str,
        description: str,
        committer: User,
    ) -> None:
        self.commit_id = commit_id
        self.parent_commit_id = parent_commit_id
        self.title = title
        self.description = description
        self.committer = committer

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.commit_id}")'

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`Commit` instance for the given contents.

        Arguments:
            contents: A dict containing all the information of the commit::

                    {
                        "commitId": <str>
                        "parentCommitId": <str> or None
                        "title": <str>
                        "description": <str>
                        "committer": {
                            "name": <str>
                            "date": <int>
                        }
                    }

        Returns:
            A :class:`Commit` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the commit information into a dict.

        Returns:
            A dict containing all the information of the commit::

                {
                    "commitId": <str>
                    "parentCommitId": <str> or None
                    "title": <str>
                    "description": <str>
                    "committer": {
                        "name": <str>
                        "date": <int>
                    }
                }

        """
        return self._dumps()


class _NamedCommit(Commit):
    """This class defines the structure of a named commit.

    :class:`_NamedCommit` is the base class of :class:`Tag` and :class:`Branch`.

    Arguments:
        name: The name of the named commit.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.

    """

    _T = TypeVar("_T", bound="_NamedCommit")

    _repr_attrs = ("commit_id",) + Commit._repr_attrs

    name: str = attr()

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        commit_id: str,
        parent_commit_id: Optional[str],
        title: str,
        description: str,
        committer: User,
    ) -> None:
        super().__init__(commit_id, parent_commit_id, title, description, committer)
        self.name = name

    def _repr_head(self) -> str:
        return f'{self.__class__.__name__}("{self.name}")'

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`_NamedCommit` instance for the given contents.

        Arguments:
            contents: A dict containing all the information of the named commit::

                    {
                        "name": <str>
                        "commitId": <str>
                        "parentCommitId": <str> or None
                        "title": <str>
                        "description": <str>
                        "committer": {
                            "name": <str>
                            "date": <int>
                        }
                    }

        Returns:
            A :class:`_NamedCommit` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the named commit information into a dict.

        Returns:
            A dict containing all the information of the named commit::

                {
                    "name": <str>
                    "commitId": <str>
                    "parentCommitId": <str> or None
                    "title": <str>
                    "description": <str>
                    "committer": {
                        "name": <str>
                        "date": <int>
                    }
                }

        """
        return self._dumps()


class Tag(_NamedCommit):
    """This class defines the structure of the tag of a commit.

    Arguments:
        name: The name of the tag.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.

    """


_ERROR_MESSAGE = "The '{attr_name}' is not available due to this branch has no commit history."
_attr = partial(attr, is_dynamic=True, error_message=_ERROR_MESSAGE)


class Branch(_NamedCommit):
    """This class defines the structure of a branch.

    Arguments:
        name: The name of the branch.
        commit_id: The commit id.
        parent_commit_id: The parent commit id.
        title: The commit title.
        description: The commit description.
        committer: The commit user.

    """

    parent_commit_id: Optional[str] = _attr(key=camel)
    title: str = _attr()
    description: str = _attr()
    committer: User = _attr()

    def _loads(self, contents: Dict[str, Any]) -> None:
        self.name = contents["name"]
        self.commit_id = contents["commitId"]

        if self.commit_id == ROOT_COMMIT_ID:
            return

        self.parent_commit_id = contents["parentCommitId"]
        self.title = contents["title"]
        self.description = contents["description"]
        self.committer = User.loads(contents["committer"])


class Draft(AttrsMixin, ReprMixin):  # pylint: disable=too-many-instance-attributes
    """This class defines the basic structure of a draft.

    Arguments:
        number: The number of the draft.
        title: The title of the draft.
        branch_name: The branch name.
        status: The status of the draft.
        parent_commit_id: The parent commit id.
        author: The author of the draft.
        updated_at: The time of last update.
        description: The draft description.

    """

    _T = TypeVar("_T", bound="Draft")

    _repr_attrs = ("title", "status", "branch_name")

    number: int = attr()
    title: str = attr()
    branch_name: str = attr(key=camel)
    status: str = attr()
    parent_commit_id: str = attr(key=camel)
    author: User = attr()
    updated_at: int = attr(key=camel)
    description: str = attr(default="")

    def __init__(  # pylint: disable=too-many-arguments
        self,
        number: int,
        title: str,
        branch_name: str,
        status: str,
        parent_commit_id: str,
        author: User,
        updated_at: int,
        description: str = "",
    ) -> None:
        self.number = number
        self.title = title
        self.branch_name = branch_name
        self.status = status
        self.parent_commit_id = parent_commit_id
        self.author = author
        self.updated_at = updated_at
        self.description = description

    def _repr_head(self) -> str:
        return f"{self.__class__.__name__}({self.number})"

    @classmethod
    def loads(cls: Type[_T], contents: Dict[str, Any]) -> _T:
        """Loads a :class:`Draft` instance from the given contents.

        Arguments:
            contents: A dict containing all the information of the draft::

                    {
                        "number": <int>
                        "title": <str>
                        "branchName": <str>
                        "status": "OPEN", "CLOSED" or "COMMITTED"
                        "parentCommitId": <str>
                        "author": {
                            "name": <str>
                            "date": <int>
                        }
                        "updatedAt": <int>
                        "description": <str>
                    }

        Returns:
            A :class:`Draft` instance containing all the information in the given contents.

        """
        return common_loads(cls, contents)

    def dumps(self) -> Dict[str, Any]:
        """Dumps all the information of the draft into a dict.

        Returns:
            A dict containing all the information of the draft::

                {
                    "number": <int>
                    "title": <str>
                    "branchName": <str>
                    "status": "OPEN", "CLOSED" or "COMMITTED"
                    "parentCommitId": <str>
                    "author": {
                        "name": <str>
                        "date": <int>
                    }
                    "updatedAt": <int>
                    "description": <str>
                }

        """
        return self._dumps()



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""Deprecated, KwargsDeprecated, DefaultValueDeprecated and Disable.

:class:`Deprecated` is a decorator for deprecated functions.

:class:`KwargsDeprecated` is a decorator for the function which has deprecated keyword arguments.

:class:`DefaultValueDeprecated` is a decorator for the function which has deprecated argument
 default value.

:class:`Disable` is a decorator for the function which is disabled temporarily.

"""

import inspect
import warnings
from functools import wraps
from typing import Any, Callable, Optional, Tuple, TypeVar, Union

_Callable = TypeVar("_Callable", bound=Callable[..., Any])


class Deprecated:
    """A decorator for deprecated functions.

    Arguments:
        since: The version the function is deprecated.
        removed_in: The version the function will be removed in.
        substitute: The substitute function.

    """

    def __init__(
        self,
        *,
        since: str,
        removed_in: Optional[str] = None,
        substitute: Union[None, str, Callable[..., Any]] = None,
    ) -> None:
        self._since = since
        self._removed_in = removed_in
        if not substitute:
            self._substitute = None
            self._meth = None
        elif callable(substitute):
            self._substitute = substitute.__qualname__
            self._meth = f":meth:`~{substitute.__module__}.{substitute.__qualname__}`"
        else:
            self._substitute = substitute.rsplit(".", 1)[-1]
            self._meth = f":meth:`~{substitute}`"

    def __call__(self, func: _Callable) -> _Callable:
        """Wrap the decorated function by adding the deprecated message.

        Arguments:
            func: The deprecated function.

        Returns:
            The wrapped function which shows the deprecated message when calling.

        """
        messages = [f'Function "{func.__name__}" is deprecated since version {self._since}.']
        if self._removed_in:
            messages.append(f'It will be removed in version "{self._removed_in}".')

        if self._substitute:
            messages.append(f'Use "{self._substitute}" instead.')

        message = " ".join(messages)

        @wraps(func)
        def wrapper(*arg: Any, **kwargs: Any) -> Any:
            warnings.warn(message, DeprecationWarning, 2)

            return func(*arg, **kwargs)

        wrapper.__doc__ = self._update_docstring(wrapper.__doc__)

        return wrapper  # type: ignore[return-value]

    def _update_docstring(self, docstring: Optional[str]) -> str:
        insert_block = [f".. deprecated:: {self._since}"]
        if self._removed_in:
            insert_block.append(f"   Will be removed in version {self._removed_in}.")
        if self._meth:
            insert_block.append(f"   Use {self._meth} instead.")

        if not docstring:
            return "\n".join(insert_block)

        lines = docstring.splitlines()

        indent = ""
        for line in lines[1:]:
            if line:
                indent = line[: -len(line.lstrip())]
                break

        lines[1:1] = (f"{indent}{message}" for message in insert_block)
        lines.insert(1, "")

        return "\n".join(lines)


class KwargsDeprecated:
    """A decorator for the function which has deprecated keyword arguments.

    Arguments:
        keywords: The keyword arguments which need to be deprecated.
        since: The version the keyword arguments are deprecated.
        remove_in: The version the keyword arguments will be removed in.
        substitute: The substitute usage.

    """

    def __init__(
        self,
        keywords: Tuple[str, ...],
        *,
        since: str,
        removed_in: Optional[str] = None,
        substitute: Optional[str] = None,
    ) -> None:
        self._keywords = keywords
        self._since = since
        self._removed_in = removed_in
        self._substitute = substitute

    def __call__(self, func: _Callable) -> _Callable:
        """Wrap the decorated function by adding the deprecated message.

        Arguments:
            func: The deprecated function.

        Returns:
            The wrapped function which shows the deprecated message when calling.

        """
        keywords = tuple(f'"{keyword}"' for keyword in self._keywords)
        if len(keywords) == 1:
            keyword_message = keywords[0]
            argument = "argument"
            be = "is"  # pylint: disable=invalid-name
        else:
            keyword_message = f"{' '.join(keywords[:-1])} and {keywords[-1]}"
            argument = "arguments"
            be = "are"  # pylint: disable=invalid-name

        messages = [
            (
                f'The keyword {argument}: {keyword_message} in "{func.__name__}" '
                f"{be} deprecated since version {self._since}."
            )
        ]
        if self._removed_in:
            messages.append(f'Will be removed in version "{self._removed_in}".')

        if self._substitute:
            messages.append(f'Use "{self._substitute}" instead.')

        message = " ".join(messages)

        @wraps(func)
        def wrapper(*arg: Any, **kwargs: Any) -> Any:
            if set(self._keywords) & kwargs.keys():
                warnings.warn(message, DeprecationWarning, 2)

            return func(*arg, **kwargs)

        return wrapper  # type: ignore[return-value]


class DefaultValueDeprecated:
    """A decorator for the function which has deprecated argument default value.

    Arguments:
        keyword: The argument keyword whose default value needs to be deprecated.
        since: The version the keyword arguments are deprecated.
        remove_in: The version the keyword arguments will be removed in.

    """

    def __init__(
        self,
        keyword: str,
        *,
        since: str,
        removed_in: Optional[str] = None,
    ) -> None:
        self._keyword = keyword
        self._since = since
        self._removed_in = removed_in

    def __call__(self, func: _Callable) -> _Callable:
        """Wrap the decorated function by adding the deprecated message.

        Arguments:
            func: The deprecated function.

        Returns:
            The wrapped function which shows the deprecated message when calling.

        """
        signature = inspect.signature(func)
        default_value = signature.parameters[self._keyword].default

        messages = [
            (
                f'The argument "{self._keyword}" in "{func.__name__}" is required '
                f'since version "{self._since}".'
            )
        ]
        if self._removed_in:
            messages.append(
                (
                    "Its default value is deprecated and "
                    f'will be removed in version "{self._removed_in}".'
                )
            )

        message = " ".join(messages)

        @wraps(func)
        def wrapper(*arg: Any, **kwargs: Any) -> Any:
            bound_argument = signature.bind(*arg, **kwargs)
            bound_argument.apply_defaults()
            if bound_argument.arguments[self._keyword] == default_value:
                warnings.warn(message, DeprecationWarning, 2)

            return func(*arg, **kwargs)

        return wrapper  # type: ignore[return-value]


class Disable:
    """A decorator for the function which is disabled temporarily.

    Arguments:
        since: The version the function is disabled temporarily.
        enabled_in: The version the function will be enabled again.
        reason: The reason that the function is disabled temporarily.

    """

    def __init__(
        self,
        *,
        since: str,
        enabled_in: Optional[str],
        reason: Optional[str],
    ) -> None:
        self._since = since
        self._enabled_in = enabled_in
        self._reason = reason

    def __call__(self, func: _Callable) -> _Callable:
        """Wrap the disabled function by adding the message.

        Arguments:
            func: The disabled function.

        Returns:
            The wrapped function which shows the message when calling.

        """
        messages = [f'The function "{func.__name__}" will be disabled since version {self._since}.']
        if self._reason:
            messages.append(f"The reason is {self._reason}.")

        if self._enabled_in:
            messages.append(f"Will be enabled again in version {self._enabled_in}.")

        message = " ".join(messages)

        @wraps(func)
        def wrapper(*arg: Any, **kwargs: Any) -> Any:
            raise NotImplementedError(message)

        return wrapper  # type: ignore[return-value]



#!/usr/bin/env python3
#
# Copyright 2021 Graviti. Licensed under MIT License.
#

"""HealthReport.

:class:`HealthReport` contains all the errors found through healthcheck
within :class:`~tensorbay.dataset.dataset.Dataset` or
:class:`~tensorbay.dataset.dataset.FusionDataset`, including the result of
basic checking and catalog checking.

"""


from collections import OrderedDict
from typing import Any, Callable, List, TypeVar

from tensorbay.utility import UserMutableMapping, UserSequence


class Error:
    """Base class of healthcheck errors."""

    def __str__(self) -> str:
        return ""


class _ErrorContext:
    _S = TypeVar("_S", bound="_ErrorContext")

    def __init__(self, title: str) -> None:
        self._title = title

    def __enter__(self: _S) -> _S:
        print("-------------------------------------------")
        print(self._title)
        return self

    def __exit__(self, *_: Any) -> None:
        if not self:
            print("PASS")


class _ErrorSection(UserSequence[Error]):
    def __init__(self) -> None:
        self._data: List[Error] = []

    def append(self, error: Error) -> None:
        """Append a error to the end of the error list.

        Arguments:
            error: The error object needs to be appended.

        """
        self._data.append(error)
        print(f"    {error}")


class _ErrorList(_ErrorSection, _ErrorContext):
    def __init__(self, title: str) -> None:
        _ErrorSection.__init__(self)
        _ErrorContext.__init__(self, title)

    def append(self, error: Error) -> None:
        """Append a error to the end of the error list.

        Arguments:
            error: The error object needs to be appended.

        """
        if not self:
            print("FAILED")

        self._data.append(error)
        print(f"  {error}")


_T = TypeVar("_T")


class _ErrorDict(UserMutableMapping[_T, _ErrorSection], _ErrorContext):
    def __init__(self, title: str, key_printer: Callable[[_T], str] = str) -> None:
        _ErrorContext.__init__(self, title)
        self._data = OrderedDict()
        self._key_printer = key_printer

    def __getitem__(self, key: _T) -> _ErrorSection:
        try:
            return self._data.__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key: _T) -> _ErrorSection:
        if not self:
            print("FAILED")
        print(f"  {self._key_printer(key)}:")
        value = _ErrorSection()
        self._data.__setitem__(key, value)
        return value


class HealthReport:
    """:class:`HealthReport` is the result of the healthcheck, which contains all the errors."""

    def __init__(self) -> None:
        self.basic_reports = _ErrorList("Basic checking:")
        self.subcatalog_reports: _ErrorDict[str] = _ErrorDict(
            "Subcatalog checking:", lambda key: key
        )



#!/usr/bin/env python3
#
# Copytright 2021 Graviti. Licensed under MIT License.
#
# pylint: disable=invalid-name

"""Dataloader of the SegTrack2 dataset."""

from tensorbay.opendataset.SegTrack2.loader import SegTrack2

__all__ = ["SegTrack2"]
